import time
from typing import Dict, Any, Tuple
import numba

import scipy
from scipy.spatial.transform import Rotation
from dataclasses import dataclass

import numpy as np

from dataspree.inference.core.config import ConfigFieldTypes
from dataspree.inference.core.config.configurable import Configurable
from dataspree.inference.core.decorators import updater


@dataclass
class AlignedPointCloud:
    aligned_cloud: np.ndarray
    invalid_height: np.ndarray
    original_point_cloud: np.ndarray


class ObjectCenter(Configurable):
    def __init__(self, x: int, y: int, z: int) -> None:
        Configurable.__init__(self)
        self.x = x
        self.y = y
        self.z = z
        self.pallet_center: np.ndarray = np.array([self.x, self.y, self.z]).astype(np.int16)

    @updater('x', 'y', 'z')
    def update_position(self):
        self.pallet_center = np.array([self.x, self.y, self.z]).astype(np.int16)

    def translate(self, point_cloud: np.ndarray, inline: bool = False) -> np.ndarray:
        if inline:
            np.subtract(point_cloud, self.pallet_center, out=point_cloud)
            return point_cloud

        return point_cloud - self.pallet_center

    @classmethod
    def get_config_entries(cls) -> Dict[str, Any]:
        return {
            'updatable': True,
            'name': 'center',
            'display_name': 'center',
            'type': ConfigFieldTypes.GROUP,
            'fields': [
                {
                    'name': 'x',
                    'display_name': 'x',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                    'default': 0,
                    'description': ''
                },
                {
                    'name': 'y',
                    'display_name': 'y',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                    'default': 0,
                    'description': ''
                },
                {
                    'name': 'z',
                    'display_name': 'z',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                    'default': 0,
                    'description': ''
                },
            ]
        }


class ObjectRotation(Configurable):
    """
    Compute the 3D orientation of an oriented surface.
    """

    @staticmethod
    def from_rectangle(vertex_1: np.ndarray, vertex_2: np.ndarray,
                       vertex_3: np.ndarray, vertex_4: np.ndarray) -> 'ObjectRotation':
        """
        Compute the rotation matrix in roll, pitch, yaw representation of a rotated rectangle in 3D.

        clockwise order of vertices is assumed.
        :param vertex_1: "Top Left" corner of the rectangle in 3D
        :param vertex_2: "Top Right" corner of the rectangle in 3D
        :param vertex_3: "Bottom Right" corner of the rectangle in 3D
        :param vertex_4: "Bottom Left" corner of the rectangle in 3D
        :return: An instance of ObjectRotation, that contains the computed yaw, pitch, and roll angles, and provides
                 the caller with a "rotate" function.
        """
        assert vertex_1.shape == (3,)
        assert vertex_2.shape == (3,)
        assert vertex_3.shape == (3,)
        assert vertex_4.shape == (3,)

        span_1 = vertex_1 - vertex_3
        span_2 = vertex_4 - vertex_3
        normal_vector = np.cross(span_1, span_2)

        stacked = np.stack((span_1, span_2, normal_vector))
        assert stacked.shape == (3, 3)
        assert np.all(stacked[0] == span_1)
        assert np.all(stacked[1] == span_2)
        assert np.all(stacked[2] == normal_vector)

        return ObjectRotation.from_coordinate_system(stacked)

    @staticmethod
    def from_coordinate_system(coordinate_system: np.ndarray) -> 'ObjectRotation':
        """
        Compute object rotation matrix in roll, pitch, yaw representation from a coordinate system (3x3) with
        coordinate_system[i] = base[i].

        coordinate system does not have to be perfect, ortnonormalization is performed.

        :param coordinate_system:  Coordinate system of the rotated object in space. Shape 3x3,
                                   coordinate_system[i]=base[i].
        :return: An instance of ObjectRotation, that contains the computed yaw, pitch, and roll angles, and provides
                 the caller with a "rotate" function.
        """
        assert coordinate_system.shape == (3, 3)

        # Orthonormalization, i.e., computation of
        #  1) unit vectors (|v|_2 = 1), that are
        #  2) orthogonal (i.e., v_i x v_j = \delta_{ij})
        # TODO: @Clemens: Here, I made a copy of the coordinate OG system to be able to check that
        #                 original coordinate system[i] is "similar" to ortonormalized coordinate system[i].
        #                 In my own implementation of orthonormalization, I could even check that
        #                       ogc[0] \approx coordinate_system[0] * | ogc[0]|_2
        #                 because we don't change the direction of the first vector.
        #       Feel free to remove this once you don't need it anymore to investigate scipy's implementation.
        ogc = coordinate_system.copy()
        coordinate_system = ObjectRotation.__orthonormalize(coordinate_system)

        # After successful orthonormalization, rotation_matrix should contain a base, and therefore constitute a
        # rotation matrix
        rotation_matrix = coordinate_system.T
        if __debug__:
            approximate_unit = rotation_matrix.dot(rotation_matrix.T)
            assert np.all(np.abs(approximate_unit - np.eye(3)) < 1e-6), \
                'The rotation matrix is not a rotation matrix, it does not seem to be orthonormal!'

        # Compute the roll pitch yaw representation of the rotation matrix
        rot = Rotation.from_matrix(rotation_matrix)

        roll, pitch, yaw = rot.as_euler('xyz', degrees=True)

        return ObjectRotation(roll=roll, pitch=pitch, yaw=yaw)

    def __init__(self, roll: float, pitch: float, yaw: float) -> None:
        Configurable.__init__(self)
        self.roll: float = roll
        self.pitch: float = pitch
        self.yaw: float = yaw

        self.rotation_matrix: np.ndarray = np.array([self.yaw, self.pitch, self.roll])
        self.r = Rotation.from_euler('zyx', -1 * self.rotation_matrix, degrees=True)
        self.r_matrix = self.r.as_matrix()

    @updater('roll', 'pitch', 'yaw')
    def update_angle(self) -> None:
        self.rotation_matrix = np.array([self.yaw, self.pitch, self.roll])
        self.r = Rotation.from_euler('zyx', -1 * self.rotation_matrix, degrees=True)
        self.r_matrix = self.r.as_matrix()

    def rotate(self, points: np.ndarray) -> np.ndarray:
        """
        Rotate and return a copy of points according the rotation stored in this instance.
        :param points: (const) points to be rotated.
        :param negate: if true, rotate against self.roll, self.pitch and self.yaw to mitigate the computed rotation.
        :return: the rotated points.
        """
        assert points.shape == (points.shape[0], points.shape[1], 3)
        pc_h, pc_w = points.shape[:-1]

        # rotated_points = self.r.apply(points.reshape(pc_h * pc_w, -1)).astype(np.int16)
        rotated_points = self.r.apply(points.reshape(pc_h * pc_w, -1).astype(float)).astype(np.int16)
        rotated_points = rotated_points.reshape((pc_h, pc_w, -1))
        return rotated_points

    @staticmethod
    def __project(vec: np.ndarray, onto: np.ndarray):
        vec = vec.astype(float)
        onto = onto.astype(float)
        return onto * vec.dot(onto) / onto.dot(onto)

    @staticmethod
    def __orthonormalize(mat: np.ndarray) -> np.ndarray:
        def gram_schmidt(A):
            """
            Here it is row wise gram schmidt. The result is a matrix Q with orthonormal rows.
            """
            Q = np.zeros_like(A)
            for i in range(A.shape[0]):
                v = A[i]
                for j in range(i):
                    r = np.dot(Q[j], A[i])
                    v -= r * Q[j]
                Q[i] = v / np.linalg.norm(v)
            return Q

        def gram_schmidt_by_julius(m2):
            assert m2.shape == (3, 3)

            m2[0] = m2[0] / np.linalg.norm(m2[0])

            # subtract the projection into the direction of the first base
            m2[1] = m2[1] - ObjectRotation.__project(vec=m2[1], onto=m2[0])
            m2[1] = m2[1] / np.linalg.norm(m2[1])

            # subtract the projection into the direction of the first & second base
            m2[2] = m2[2] - \
                     ObjectRotation.__project(vec=m2[2], onto=m2[0]) - \
                     ObjectRotation.__project(vec=m2[2], onto=m2[1])
            m2[2] = m2[2] / np.linalg.norm(m2[2])
            return m2

        def symmetric_using_svd(A):
            return scipy.linalg.orth(A)

        # mat = np.array([[2, 0, 0], [0, 5, 0], [0, 0, 7]])
        mat = mat.astype(float, copy=True)
        a = np.linalg.qr(mat.copy().T)[0]
        b = gram_schmidt(mat.copy())
        c = gram_schmidt_by_julius(mat.copy())
        d = np.flipud(symmetric_using_svd(mat.copy().T))

        return b

    @classmethod
    def get_config_entries(cls) -> Dict[str, Any]:
        return {
            'updatable': True,
            'name': 'rotation',
            'display_name': 'rotation',
            'type': ConfigFieldTypes.GROUP,
            'fields': [
                {
                    'name': 'roll',
                    'display_name': 'roll',
                    'type': ConfigFieldTypes.DOUBLE_INPUT,
                    'default': 3.7,
                    'description': ''
                },
                {
                    'name': 'pitch',
                    'display_name': 'pitch',
                    'type': ConfigFieldTypes.DOUBLE_INPUT,
                    'default': -13.0,
                    'description': ''
                },
                {
                    'name': 'yaw',
                    'display_name': 'yaw',
                    'type': ConfigFieldTypes.DOUBLE_INPUT,
                    'default': 0.0,
                    'description': ''
                },
            ]
        }


class ObjectPosition(Configurable):

    def __init__(self, rotation: ObjectRotation, center: ObjectCenter) -> None:
        Configurable.__init__(self)
        self.rotation = rotation
        self.center = center

        # self.timings = np.zeros((4, 50))
        # self.iteration = 0

    def align(self, point_cloud: np.ndarray, negate_z: bool = False) -> AlignedPointCloud:
        return self.__align_fast(point_cloud, negate_z)

    def __align_fast(self, point_cloud: np.ndarray, negate_z) -> AlignedPointCloud:
        """
        This function is an order of magnitudes faster than the more readable  align_slow function
        """

        # times = [time.perf_counter_ns()]
        aligned_point_cloud, invalid = align_fast(point_cloud, self.rotation.r_matrix, self.center.pallet_center,
                                                  negate_z)
        # times.append(time.perf_counter_ns())

        ap = AlignedPointCloud(
            aligned_cloud=aligned_point_cloud,
            invalid_height=invalid,
            original_point_cloud=point_cloud
        )
        # times.append(time.perf_counter_ns())

        # self.timings[:len(times) - 1, self.iteration] = [(t2 - t1) // 1e6. for t2, t1 in zip(times[1:], times[:-1])]
        # self.iteration = (self.iteration + 1) % self.timings.shape[1]
        # if self.iteration == 0:
        #    print(np.mean(self.timings, axis=1))

        return ap

    def __align_slow(self, point_cloud: np.ndarray) -> AlignedPointCloud:
        # times = [time.perf_counter_ns()]
        rotated_point_cloud: np.ndarray = self.rotation.rotate(point_cloud)
        # times.append(time.perf_counter_ns())

        offset_and_rotated_point_cloud: np.ndarray = self.center.translate(rotated_point_cloud, inline=True)
        # times.append(time.perf_counter_ns())

        invalid: np.ndarray = point_cloud[..., 2] == 0

        # Remove invalid pixels (not auto removed if translation is non-trivial).
        offset_and_rotated_point_cloud[invalid] = 0
        # times.append(time.perf_counter_ns())

        ap = AlignedPointCloud(
            aligned_cloud=offset_and_rotated_point_cloud,
            invalid_height=invalid,
            original_point_cloud=point_cloud
        )
        # times.append(time.perf_counter_ns())
        # self.timings[..., self.iteration] = [(t2 - t1) // 1e6 for t2, t1 in zip(times[1:], times[:-1])]
        # self.iteration = (self.iteration + 1) % self.timings.shape[1]
        # if self.iteration == 0:
        #    print(np.mean(self.timings, axis=1))

        return ap

    @classmethod
    def get_config_entries(cls, key: str = 'object') -> Dict[str, Any]:
        return {
            'updatable': True,
            'name': key,
            'display_name': key.capitalize(),
            'type': ConfigFieldTypes.GROUP,
            'fields': [
                ObjectCenter.get_config_entries(),
                ObjectRotation.get_config_entries()
            ]
        }


@numba.jit(numba.types.Tuple((numba.types.int16[:, :, :], numba.types.boolean[:, :]))(
    numba.types.int16[:, :, :], numba.types.float64[:, :], numba.types.int16[:], numba.types.boolean), nopython=True)
def align_fast(point_cloud: np.ndarray, rotation: np.ndarray, center: np.ndarray, negate_z: bool) \
        -> Tuple[np.ndarray, np.ndarray]:
    invalid: np.ndarray = np.zeros((point_cloud.shape[0], point_cloud.shape[1]), dtype=np.bool_)
    aligned_point_cloud: np.ndarray = np.zeros_like(point_cloud)
    for i in range(point_cloud.shape[0]):
        for j in range(point_cloud.shape[1]):
            if point_cloud[i, j, -1] == 0:
                invalid[i, j] = True
            else:
                p = point_cloud[i, j]
                if negate_z:
                    aligned_point_cloud[i, j] = (
                        rotation[0, 0] * p[0] + rotation[0, 1] * p[1] + rotation[0, 2] * p[2] - center[0],
                        rotation[1, 0] * p[0] + rotation[1, 1] * p[1] + rotation[1, 2] * p[2] - center[1],
                        -rotation[2, 0] * p[0] - rotation[2, 1] * p[1] - rotation[2, 2] * p[2] + center[2])
                else:
                    aligned_point_cloud[i, j] = (
                        rotation[0, 0] * p[0] + rotation[0, 1] * p[1] + rotation[0, 2] * p[2] - center[0],
                        rotation[1, 0] * p[0] + rotation[1, 1] * p[1] + rotation[1, 2] * p[2] - center[1],
                        rotation[2, 0] * p[0] + rotation[2, 1] * p[1] + rotation[2, 2] * p[2] - center[2],)



    return aligned_point_cloud, invalid
