import numpy as np
from dataspree.inference.core.decorators import numba_fallback, numba


@numba_fallback
def get_display_image(intensity_image: np.ndarray):
    height, width = intensity_image.shape[:2]
    center_patch = intensity_image[int(height * 0.4):int(height * 0.6), int(width * 0.4):int(width * 0.6)]

    norm_min = center_patch.min() * 0.95
    norm_max = center_patch.max() * 1.05
    amp_img = intensity_image[..., 0].astype(np.float16, copy=True)
    np.clip(amp_img, norm_min, norm_max, amp_img)
    np.subtract(amp_img, norm_min, out=amp_img)
    if norm_min != norm_max:
        np.multiply(amp_img, 255.0 / (norm_max - norm_min), out=amp_img)

    return amp_img


@get_display_image.register_numba_implementation
@numba.jit(numba.uint8[:, :](numba.uint16[:, :, :]), nopython=True)
def get_display_image_numba(intensity_image: np.ndarray):
    height, width = intensity_image.shape[:2]

    center_patch = intensity_image[int(height * 0.4):int(height * 0.6), int(width * 0.4):int(width * 0.6)]
    v_min, v_max = np.iinfo(intensity_image.dtype).min, np.iinfo(intensity_image.dtype).max
    for row in range(center_patch.shape[0]):
        for col in range(center_patch.shape[1]):
            if center_patch[row, col, 0] < v_min:
                v_min = center_patch[row, col, 0]
            if center_patch[row, col, 0] > v_max:
                v_max = center_patch[row, col, 0]

    v_min *= .95
    v_max *= 1.05
    diff = v_max - v_min
    display_image = np.empty(intensity_image.shape[:2], dtype=np.uint8)
    for row in range(height):
        for col in range(width):
            val = min(max(intensity_image[row, col, 0], v_min), v_max)
            display_image[row, col] = (val - v_min) / diff * 255

    return display_image


@numba_fallback
def process_point_cloud(distance_image: np.ndarray, transformation_matrix: np.ndarray,
                        s0_matrix: np.ndarray, f2rc: int):
    pc = transformation_matrix * distance_image
    np.divide(pc, s0_matrix[..., None], out=pc, where=s0_matrix[..., None] != 0)
    np.subtract(pc[..., 2], f2rc, out=pc[..., 2])
    return pc.astype(np.int32)


@process_point_cloud.register_numba_implementation
@numba.jit(numba.int32[:, :, :](numba.float64[:, :, :], numba.float32[:, :, :], numba.float32[:, :], numba.float64),
           nopython=True, fastmath=True)
def process_point_cloud_numba(distance_image: np.ndarray, transformation_matrix: np.ndarray,
                              s0_matrix: np.ndarray, f2rc: int) -> np.ndarray:
    pc = np.empty_like(transformation_matrix, dtype=np.int32)
    for row in range(distance_image.shape[0]):
        for col in range(distance_image.shape[1]):
            pc[row, col] = (transformation_matrix[row, col] * distance_image[row, col]) / s0_matrix[row, col]
            pc[row, col, 2] -= f2rc

    return pc
