from albumentations import Crop
from typing import Optional, List, Set, Dict
from dataclasses import dataclass, field
import random

from itertools import product


class RelativeCrop(Crop):

    def __init__(self, x_min_percent: float = 0.0, y_min_percent: float = 0.0,
                 x_max_percent: float = 1.0, y_max_percent=1.0, always_apply=False, p=1.0):
        super().__init__(x_min=0, y_min=0, x_max=1, y_max=1, always_apply=always_apply, p=p)

        # In development mode ensure compatibility with the implementation of crop shipped with the version of
        # Albumentation that is currently in-use.
        if any((getattr(self, var, None) is None for var in ('x_min', 'x_max', 'y_min', 'y_max'))):
            raise NotImplementedError(
                'This version of Albumentations is incompatible with the RelativeCrop augmentation.'
                'Make sure that you use the recommended version of Albumentations and report'
                'this incompatibility to a DataSpree developer.')

        self.x_min_percent: float = x_min_percent
        self.y_min_percent: float = y_min_percent
        self.x_max_percent: float = x_max_percent
        self.y_max_percent: float = y_max_percent

        if not (0 <= self.x_min_percent <= self.x_max_percent <= 1):
            raise ValueError(f'The x coordinates for the Transformation "{type(self).__name__}" are invalid:'
                             f'"{0} <= {self.x_min_percent} = x min [%] <= {self.x_max_percent} = x max [%] <= 1" '
                             f'does not hold')

        if not (0 <= self.y_min_percent <= self.y_max_percent <= 1):
            raise ValueError(f'The y coordinates for the Transformation "{type(self).__name__}" are invalid:'
                             f'"{0} <= {self.y_min_percent} = y min [%] <= {self.y_max_percent} = y max [%] <= 1" '
                             f'does not hold')
        self.__last_input_width: int = 0
        self.__last_input_height: int = 0
        self.applied_in_replay = True #XXX

    def update_params(self, params, **kwargs):
        # @dev I think update_params is the right choice here, if not check out & implement
        # def get_params_dependent_on_targets(self, params: Dict[str, Any]) -> Dict[str, Any]:

        params = super(RelativeCrop, self).update_params(params, **kwargs)
        self.__last_input_height = params['rows']
        self.__last_input_width = params['cols']
        self.update_child()
        return super(RelativeCrop, self).update_params(params, **kwargs)

    def update_child(self):
        self.x_min = round(self.x_min_percent * self.__last_input_width)
        self.y_min = round(self.y_min_percent * self.__last_input_height)
        self.x_max = max(round(self.x_max_percent * self.__last_input_width), self.x_min + 1)
        self.y_max = max(round(self.y_max_percent * self.__last_input_height), self.y_min + 1)

        # XXX: kriegen wir das hier noch schöner hin?
        if self.y_max > self.__last_input_height:
            self.y_max -= 1
            self.y_min -= 1

        if self.x_max > self.__last_input_width:
            self.x_max -= 1
            self.x_min -= 1

    def __check_after_update(self, rows: Optional[int] = None, cols: Optional[int] = None, **_kwargs):
        if rows is None and cols is None:
            raise NotImplementedError('Rows and Columns have to be provided by a compatible version of Albumentations. '
                                      'Make sure you use the recommended version of Albumentations and report this '
                                      'incompatibility to a DataSpree developer.')
        else:
            if rows != self.__last_input_height:
                raise NotImplementedError(f'Rows {rows} does not match expected height {self.__last_input_height}. '
                                          f'The current version of Albumentations is incompatible with the '
                                          f'{type(self).__name__} augmentation.'
                                          'Make sure that you use the recommended version of Albumentations and report '
                                          'this incompatibility to a DataSpree developer.')
            if cols != self.__last_input_width:
                raise NotImplementedError(f'Rows {cols} does not match expected height {self.__last_input_width}'
                                          f'The current version of Albumentations is incompatible with the '
                                          f'{type(self).__name__} augmentation.'
                                          'Make sure that you use the recommended version of Albumentations and report '
                                          'this incompatibility to a DataSpree developer.')

    def apply(self, img, **params):
        self.__check_after_update(**params)
        return super(RelativeCrop, self).apply(img, **params)

    def apply_to_bbox(self, bbox, **kwargs):
        self.__check_after_update(**kwargs)
        return super(RelativeCrop, self).apply_to_bbox(bbox, **kwargs)

    def apply_to_keypoints(self, keypoints, **params):
        self.__check_after_update(**params)
        return super(RelativeCrop, self).apply_to_keypoints(keypoints, **params)

    def get_transform_init_args_names(self):
        # @dev: One could also argue not overriding this, and allowing the relative crop to optionally receive
        #       x_min, y_min, x_max, y_max (I think this is relevant for replay, and replaying an augmentation
        #       should probably require the input shape not to change.
        return 'x_min_percent', 'y_min_percent', 'x_max_percent', 'y_max_percent'

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__


"""
@dataclass(frozen=True)
class CropData:
    index: int
    max: int
    x1: float
    x2: float
    y1: float
    y2: float
"""


class CropData:
    def __init__(self, index: int, max: int, x1: float, x2: float, y1: float, y2: float):
        self.index = index
        self.max = max
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2


# @dataclass
class CropsData:

    def __init__(self, data: List[CropData]):
        self.data: List[CropData] = data
        self.__post_init__()

    def __getitem__(self, idx):
        return self.data.__getitem__(idx)

    def __len__(self):
        return len(self.data)

    def __post_init__(self) -> None:
        """
        Build two-dimensional index for crops.
        """
        x_values = []
        y_values = []
        for crop in self.data:
            x_values.append(crop.x1)
            x_values.append(crop.x2)
            y_values.append(crop.y1)
            y_values.append(crop.y2)

        self.x_values: List[float] = sorted(x_values)
        self.y_values: List[float] = sorted(y_values)

        self.x_index: Dict[float, Set[CropData]] = {x: set() for x in x_values}
        self.y_index: Dict[float, Set[CropData]] = {y: set() for y in y_values}

        for crop in self.data:
            for x in x_values:
                if x >= crop.x2:
                    break

                if crop.x1 <= x < crop.x2:
                    self.x_index[x].add(crop)

            for y in y_values:
                if y >= crop.y2:
                    break

                if crop.y1 <= y < crop.y2:
                    self.y_index[y].add(crop)

    def find_crop(self, x: float, y: float) -> Set[CropData]:
        """
        Find the set of crops that contain the given point.
        """
        crops_x = self.x_index.get(self._find(self.x_values, x), set())
        crops_y = self.y_index.get(self._find(self.y_values, y), set())
        return crops_x & crops_y

    @staticmethod
    def _find(values, lookup):
        # Find last value that is smaller than lookup.
        prev = None
        for value in values:
            if value > lookup:
                return prev
            prev = value
        return prev


class MeshCrop(RelativeCrop):

    def __init__(self, number_crops_x: int, number_crops_y: int,
                 relative_crop_width: float, relative_crop_height: float, **_kwargs) -> None:
        super().__init__(x_min_percent=0.0, y_min_percent=0.0, x_max_percent=1.0, y_max_percent=1.0,
                         always_apply=True, p=1.0)
        self.number_crops_x = number_crops_x
        self.number_crops_y = number_crops_y
        self.relative_crop_width = relative_crop_width
        self.relative_crop_height = relative_crop_height
        self.crop: Optional[CropData] = None
        self.crops: Optional[CropsData] = None
        self.all_crops = CropsData([CropData(
            i, self.batches_per_index, self.x1(ix), self.x2(ix), self.y1(iy), self.y2(iy))
            for i, (ix, iy) in enumerate(product(range(number_crops_x), range(number_crops_y)))
        ])
        self.params = {}

    def apply(self, img, **params):
        result = super().apply(img, **params)
        return result

    def apply_to_keypoints(self, keypoints, rows: Optional[int] = None, cols: Optional[int] = None, **params):
        if self.crops is not None and len(self.crops.data):
            selected_crop = None

            if len(self.crops.data) > 1:
                # Choice A:
                # Chose random detection that is inside a crop.
                # This way, we would sample crops with more detections more often.
                # Attention: Detections don't have to be contained in any crop. Resample if not found.

                # Choice B:
                # Sample uniformly from all crops containing samples.
                centers = keypoints[::3]
                box_centers = centers[:-1]
                crops_with_samples = set()
                for box in box_centers:
                    x_px, y_px = box[:2]
                    y_percent = y_px / rows  # height ~ rows ~ y
                    x_percent = x_px / cols
                    crops_with_samples |= self.crops.find_crop(x_percent, y_percent)

                if len(crops_with_samples):
                    selected_crop = random.choice(list(crops_with_samples))
                else:
                    if len(box_centers):
                        print('weird')

            if selected_crop is None:
                selected_crop = random.choice(self.crops.data)

            self.x_min_percent = selected_crop.x1
            self.x_max_percent = selected_crop.x2
            self.y_min_percent = selected_crop.y1
            self.y_max_percent = selected_crop.y2
            self.crop = selected_crop
            self.update_child()

        return super().apply_to_keypoints(keypoints, rows=rows, cols=cols, **params)

    def update_crops(self, crops: CropsData):
        self.always_apply = True
        self.p = 1.0
        self.crops = crops
        self.applied_in_replay = True
        self.update_child()

    def apply_crop(self, crop: CropData):
        self.crops = CropsData([crop])
        self.always_apply = True
        self.applied_in_replay = True
        self.p = 1.0
        self.update_child()

    def get_transform_init_args_names(self):
        return 'number_crops_x', 'number_crops_y', 'relative_crop_width', 'relative_crop_height'

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__

    def x1(self, i) -> float:
        if not (0 <= i < self.number_crops_x):
            raise Exception(f'i={i} must be in the range [0, {self.number_crops_x}).')

        x = i * (1 - self.relative_crop_width)
        if x == 0:
            return 0

        assert self.number_crops_x > 1, ('Number of crops is larger than 1, '
                                         'because otherwise i would be 0 and we would have already returned.')
        return x / (self.number_crops_x - 1)

    def x2(self, i) -> float:
        return self.x1(i) + self.relative_crop_width

    def y1(self, i) -> float:
        if not (0 <= i < self.number_crops_y):
            raise Exception(f'i={i} must be in the range [0, {self.number_crops_y}).')

        x = i * (1 - self.relative_crop_height)
        if x == 0:
            return 0

        assert self.number_crops_y > 1, ('Number of crops is larger than 1, '
                                         'because otherwise i would be 0 and we would have already returned.')
        return x / (self.number_crops_y - 1)

    def y2(self, i) -> float:
        return self.y1(i) + self.relative_crop_height

    @property
    def batches_per_index(self) -> int:
        return self.number_crops_x * self.number_crops_y
