"""This module defines the comprehensive class that represents a camera model."""

from dataclasses import dataclass, field

from typing import Union, Optional

from dataspree.inference.sensor.data.pixel_format import PixelFormat
from dataspree.inference.sensor.data.sensor_model import SensorModel


@dataclass
class CameraModel(SensorModel):
    """Representation of the camera model.

    Attributes:
        available_pixel_formats (List[PixelFormat]):
            Supported pixel formats.

        available_configuration_names (List[str])
            The names of writable configurations.

        available_configuration_names_read_only (List[str])
            The names of factory settings that cannot be adjusted / saved to.

        acquisition_burst_modes (List[str]):
            Modes for acquisition bursts.

        white_balance_manual_channels (List[Union[str, int]]):
            Channels for manual white balance adjustment.

        rois (List[str]):
            Regions of interest.

        auto_function_profiles (List[str]):
            Profiles for automatic functions.

        gpio_lines (List[str]):
            General-purpose input/output lines.

        software_trigger_sources_generic (List[str]):
            Generic software trigger sources.

        software_trigger_frame_start (Optional[str]):
            Software trigger for frame start.

        available_channels (List[str]):
            Available channels.

        binning_modes (List[str]):
            Modes for binning.

        contrast_modes (List[str]):
            Modes for contrast adjustment.

        supported_denoising_modes (Optional[List[str]]):
            Supported denoising modes.

        supported_demosaicing_modes (List[str]):
            Supported demosaicing modes.

        frame_start_trigger (Optional[str]):
            Trigger for frame start.

        frame_stop_trigger (Optional[str]):
            Trigger for frame stop.

        frame_active_trigger (Optional[str]):
            Trigger for frame active.

        frame_burst_start_trigger (Optional[str]):
            Trigger for frame burst start.

        frame_burst_stop_trigger (Optional[str]):
            Trigger for frame burst stop.

        frame_burst_active_trigger (Optional[str]):
            Trigger for frame burst active.

        exposure_start_trigger (Optional[str]):
            Trigger for exposure start.

        exposure_stop_trigger (Optional[str]):
            Trigger for exposure stop.

        exposure_active_trigger (Optional[str]):
            Trigger for exposure active.

        line_start_trigger (Optional[str]):
            Trigger for line start.




    """

    available_pixel_formats: list[PixelFormat] = field(default_factory=list)
    available_configuration_names_writable: list[str] = field(default_factory=list)
    available_configuration_names_read_only: list[str] = field(default_factory=list)
    available_configuration_names: list[str] = field(default_factory=list)

    acquisition_burst_modes: list[str] = field(default_factory=list)

    white_balance_manual_channels: list[Union[str, int]] = field(default_factory=list)
    rois: list[str] = field(default_factory=list)
    auto_function_profiles: list[str] = field(default_factory=list)

    gpio_lines: list[str] = field(default_factory=list)

    software_trigger_sources_generic: list[str] = field(default_factory=list)
    software_trigger_frame_start: Optional[str] = 'Software'

    available_channels: list[str] = field(default_factory=list)
    binning_modes: list[str] = field(default_factory=list)
    contrast_modes: list[str] = field(default_factory=list)

    supported_denoising_modes: Optional[list[str]] = field(default_factory=list)
    supported_demosaicing_modes: list[str] = field(default_factory=list)

    # TODO: Update through image_acquirer.remote_device.node_map.TriggerSelector.symbolics
    # TODO: DeviceScanType contains information on whether this is a lince scanner or area scan device.
    frame_start_trigger: Optional[str] = None
    frame_stop_trigger: Optional[str] = None
    frame_active_trigger: Optional[str] = None

    frame_burst_start_trigger: Optional[str] = None
    frame_burst_stop_trigger: Optional[str] = None
    frame_burst_active_trigger: Optional[str] = None

    exposure_start_trigger: Optional[str] = None
    exposure_stop_trigger: Optional[str] = None
    exposure_active_trigger: Optional[str] = None

    line_start_trigger: Optional[str] = None

    def __post_init__(self) -> None:
        """Initializes the `available_configuration_names` attribute.

        If the `available_configuration_names` attribute is empty, it combines the writable
        and read-only configuration names to populate it.
        """
        if not len(self.available_configuration_names):
            self.available_configuration_names = (
                self.available_configuration_names_writable + self.available_configuration_names_read_only
            )

    @property
    def trigger_sources_generic(self) -> list[str]:
        """Gets all generic trigger sources.

        Combines GPIO lines and software trigger sources.

        Returns:
            list[str]: A list of generic trigger source names.
        """
        return self.gpio_lines + self.software_trigger_sources_generic

    @property
    def all_trigger_names(self) -> set[str]:
        """Gets all unique trigger names for acquisition and exposure.

        Returns:
            set[str]: A set of all trigger names.
        """
        return self.acquisition_trigger_names | self.exposure_trigger_names

    @property
    def acquisition_trigger_names(self) -> set[str]:
        """Gets all unique trigger names related to acquisition.

        Returns:
            set[str]: A set of acquisition-related trigger names.
        """
        return {
            t
            for t in (
                self.frame_start_trigger,
                self.frame_stop_trigger,
                self.frame_active_trigger,
                self.frame_burst_start_trigger,
                self.frame_burst_stop_trigger,
                self.frame_burst_active_trigger,
            )
            if t is not None
        }

    @property
    def exposure_trigger_names(self) -> set[str]:
        """Gets all unique trigger names related to exposure.

        Returns:
            set[str]: A set of exposure-related trigger names.
        """
        return {
            e
            for e in (self.exposure_start_trigger, self.exposure_stop_trigger, self.exposure_active_trigger)
            if e is not None
        }
