"""Module for managing optional configurations and their availability states.

This module defines the `OptionalConfiguration` class and related subclasses for managing configurations that may
or may not be available. It supports structured handling of various states of availability, such as not retrieved,
not specified, and temporarily unavailable configurations.

The `OptionalConfiguration` class provides methods to determine if one configuration entails another, as well as
utility functions (`is_available` and `is_deactivated`) for checking the status of configuration values.

Several subclasses represent specific availability states for configurations, for detailed handling of availability
in configuration contexts.
"""

# TODO: Document the intended behavior and unit test it!
from typing import Any, Union, Optional, TypeGuard, TypeVar


class OptionalConfiguration:
    """Represents a configuration that may or may not be available and supports availability checking."""

    def __init__(self) -> None:
        """Initializes the configuration with an initialized state set to True."""
        self.__initialized: bool = True

    def entails(self, other: 'OptionalConfiguration') -> bool:
        """Checks if this configuration encompasses the properties of another configuration.

        Args:
            other (OptionalConfiguration): The configuration to compare against.

        Returns:
            bool: True if this configuration encompasses the other, False otherwise.
        """
        if not is_available(other):
            return True

        if not is_available(self):
            return False

        if type(self) is not type(other):
            return False

        for a in vars(self):
            try:
                mine = getattr(self, a)
                others = getattr(other, a)
            except AttributeError:
                return False
            if not is_available(others):
                continue

            try:
                if not mine.entails(others):
                    return False
            except AttributeError:
                if others != mine:
                    return False
        return True

    def __setattr__(self, key: str, value: Any) -> None:
        """Sets an attribute value, ensuring configuration validity based on availability rules.

        Args:
            key (str): The attribute name.
            value (Any): The value to set for the attribute.
        """
        try:
            initialized = self.__initialized
        except AttributeError:
            initialized = False

        if not initialized:
            super().__setattr__(key, value)
            return

        else:
            if is_available(value):
                return super().__setattr__(key, value)

            previous_value = getattr(self, key, value)
            if previous_value is None:
                return super().__setattr__(key, value)

            # previous value is not None, but might still be unavailable.
            if isinstance(
                value,
                NotSupported,
            ):
                return super().__setattr__(key, value)

            elif isinstance(value, NotRetrieved):
                pass

            elif isinstance(value, NotSpecified):
                pass


class NotAvailable(OptionalConfiguration):
    """Indicates that a configuration or feature is not available.

    To indicate that a feature is disabled, use None or the appropriate value (f.i. for a frame burst, frame burst
    would be disabled by setting the image count to 1, so it is not necessary to allow a None value.)
    """


class NotAvailableRead(NotAvailable):
    """Indicates that a configuration or feature is not available for reading."""


class NotSupported(NotAvailableRead):
    """Indicates that a feature is not supported."""


class NotRetrieved(NotAvailableRead):
    """Indicates that a configuration has not been retrieved."""


class TemporarilyNotAvailable(NotRetrieved):
    """Indicates that a configuration is temporarily unavailable."""


class NotAvailableWrite(NotAvailable):
    """Indicates that a configuration or feature is not available for writing."""


class NotSpecified(NotAvailableWrite):
    """Indicates that a configuration has not been specified."""


T = TypeVar('T')


def is_available(value: Optional[Union[T, NotAvailable]]) -> TypeGuard[T]:
    """Checks if a value is available.

    Args:
        value (Any): The value to check.

    Returns:
        bool: True if the value is available, False otherwise.
    """
    return not isinstance(value, NotAvailable) and value is not None


def is_deactivated(value: Any) -> bool:
    """Checks if a value is deactivated (set to None).

    Args:
        value (Any): The value to check.

    Returns:
        bool: True if the value is deactivated, False otherwise.
    """
    return value is None


class Auto:
    """Placeholder class for automatically set configurations."""
