"""Module for defining a structured hierarchy of exceptions for handling sensor and driver errors.

This module defines a structured hierarchy of exceptions for sensor and driver operations, supporting
error handling in scenarios where features, configurations, or resources are temporarily or permanently unavailable.

The module uses an `abstract_class` decorator to prevent instantiation of base exception classes directly,
ensuring that only concrete subclasses can be instantiated. This enforces a clear and consistent structure
for exception handling, distinguishing between sensor, driver, and configuration errors, among other categories.

Exceptions are designed to communicate both transient and persistent issues in the sensor and driver
lifecycles, including invalid states, configuration errors, and feature unavailability, providing
specific error classes for detailed feedback in each scenario.
"""

from dataspree.inference.core.exceptions import InferenceError
from typing import Any, TypeVar, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from dataspree.inference.sensor.data import Sensor, SensorState

T = TypeVar('T')


def abstract_class(cls: type[T]) -> type[T]:
    """A decorator that prevents instantiation of the decorated class directly.

    Args:
        cls (Type[T]): The class to be decorated, which will become non-instantiable directly.

    Returns:
        Type[T]: A subclassed version of the original class that raises a TypeError if instantiated directly.

    Raises:
        TypeError: If an attempt is made to instantiate the decorated class directly.

    Example:
        @abstract_class
        class MyAbstractClass:
            pass

        instance = MyAbstractClass()  # Raises TypeError
    """

    class Wrapped(cls):  # type: ignore
        """A wrapper class to enforce abstract base class behavior."""

        def __new__(cls: type['Wrapped'], *args: Any, **kwargs: Any) -> type[T]:
            """Creates a new instance of the wrapped class if not abstract.

            Raises:
                TypeError: If instantiation of the abstract class `Wrapped` is attempted.

            Returns:
                Wrapped: A new instance of the subclass, if allowed.
            """
            if cls is Wrapped:
                raise TypeError(f'Cannot instantiate abstract class {cls.__name__} directly.')
            return super().__new__(cls)

    return Wrapped


@abstract_class
class InputException(InferenceError):
    """Base class for input exceptions."""


@abstract_class
class SensorException(InputException):
    """Base class for sensor exceptions."""


@abstract_class
class DriverException(InputException):
    """Base class for driver exceptions."""


# Invalid state


@abstract_class
class InvalidState(InputException):
    """The sensor or driver is in an invalid state and therefore cannot perform the operation."""


class DriverInvalidState(InvalidState, DriverException):
    """The driver is in an invalid state and therefore cannot perform the operation."""


class SensorInvalidState(InvalidState, SensorException):
    """The sensor is in an invalid state and therefore cannot perform the operation."""

    @staticmethod
    def from_sensor(sensor: 'Sensor', expected_states: Optional[set['SensorState']] = None) -> 'SensorInvalidState':
        """Creates a SensorInvalidState exception with details about the sensor's current and expected states.

        Args:
            sensor (Sensor): The sensor in an invalid state.
            expected_states (Optional[set[SensorState]]): A set of allowed sensor states, if applicable.

        Returns:
            SensorInvalidState: An exception instance with a descriptive error message.
        """
        return SensorInvalidState(
            f'Sensor {sensor.identifier} is in invalid state {sensor.state}.'
            + ('' if not expected_states else f'Allowed states: {", ".join(map(str, expected_states))}.')
        )


class SensorNotFound(SensorInvalidState):
    """Sensor lost or not available."""


class SensorOperationNotPossibleInCurrentConfiguration(SensorInvalidState):
    """The requested operation is impossible with the current sensor configuration."""


class SensorLost(SensorNotFound):
    """Sensor lost."""

    def __init__(self, sensor: 'Sensor') -> None:
        """Initializes the SensorLost exception with the lost sensor's details.

        Args:
            sensor (Sensor): The sensor that has been lost.
        """
        super().__init__(f'Sensor {sensor} not found')
        self.sensor: Sensor = sensor


class DriverStateInconsistent(DriverInvalidState):
    """The driver state is inconsistent."""


# Configuration failed


@abstract_class
class ConfigurationError(InputException):
    """Applying the configuration failed."""


class ConfigurationInconsistent(ConfigurationError):
    """The configuration is not consistent."""


class ConfigurationValueError(ConfigurationError, ValueError):
    """Configuration options are out of range."""


class NoFreeHardwareTrigger(ConfigurationInconsistent):
    """No hardware trigger available."""

    @staticmethod
    def create() -> 'NoFreeHardwareTrigger':
        """Creates a `NoFreeHardwareTrigger` exception with a default error message.

        Returns:
            NoFreeHardwareTrigger: An instance of the exception with a pre-defined message.
        """
        return NoFreeHardwareTrigger('No free hardware trigger available.')


class NoFreeSoftwareTrigger(ConfigurationInconsistent):
    """No free software trigger available."""

    @staticmethod
    def create() -> 'NoFreeSoftwareTrigger':
        """Creates a `NoFreeSoftwareTrigger` exception with a default error message.

        Returns:
            NoFreeSoftwareTrigger: An instance of the exception with a pre-defined message.
        """
        return NoFreeSoftwareTrigger('No free software trigger available.')


class NoFreeRois(ConfigurationInconsistent):
    """No free rois available."""

    @staticmethod
    def create() -> 'NoFreeRois':
        """Creates a `NoFreeRois` exception with a default error message.

        Returns:
            NoFreeRois: An instance of the exception with a pre-defined message.
        """
        return NoFreeRois('No free regions of interest available.')


# Resource temporarily not available


@abstract_class
class ResourceError(InputException):
    """Resource not available."""


@abstract_class
class SensorResourceError(ResourceError, SensorException):
    """Resource not available on the sensor."""


class SensorResourceTemporarilyNotAvailable(SensorResourceError):
    """Resource temporarily not available."""


class SensorResourcePermanentlyNotAvailable(SensorResourceError):
    """Resource permanently (or depending on configuration) not available."""


class IncompleteBufferError(SensorResourceTemporarilyNotAvailable):
    """Could not fetch complete buffer."""


#
# Errors that convey, that a specific feature (currently) does not work
#


@abstract_class
class FeatureUnavailable(InputException):
    """The feature is temporarily or permanently not available."""


@abstract_class
class DriverFeatureUnavailable(FeatureUnavailable, DriverException):
    """The driver feature is temporarily or permanently not available."""


class SensorFeatureUnavailable(FeatureUnavailable, SensorException):
    """The sensor feature is temporarily or permanently not available."""


class DriverFeatureImplementationError(DriverFeatureUnavailable):
    """The feature is available, but in the current setting an error occurred."""


class DriverFeatureNotImplemented(DriverFeatureImplementationError):
    """The driver did not implement the feature."""


class SensorFeaturePermanentlyNotAvailable(SensorFeatureUnavailable):
    """The sensor does not provide / implement the feature."""


class SensorFeatureTemporarilyNotAvailable(SensorFeatureUnavailable):
    """The sensor is unable to execute the feature right now, but retrying might help."""
