"""Module for handling interface-specific pixel format designations.

This module provides the `InterfaceSpecific` abstract base class and its implementations
for representing and parsing interface-specific pixel format designations.
"""

from abc import ABCMeta, abstractmethod
from logging import getLogger


from dataspree.inference.sensor.data.pixel_format.number_of_bits import Size
from dataspree.inference.sensor.data.pixel_format.components import (
    channel_name,
    Components,
    LMNLocation,
    GenicamComponents,
)
from dataspree.inference.sensor.data.pixel_format.utils import all_implementations

logger = getLogger(__name__)


class InterfaceSpecific(metaclass=ABCMeta):
    """Abstract base class for interface-specific pixel format designations."""

    @abstractmethod
    def designation(self) -> str:
        """Returns the designation of the interface-specific format."""

    @staticmethod
    def create(full_designation: str, components: Components, sizes: Size) -> list[tuple['InterfaceSpecific', str]]:
        """Creates instances of `InterfaceSpecific` based on the designation.

        Args:
            full_designation (str): The full designation string.
            components (Components): The pixel format components.
            sizes (Size): The sizes of the components.

        Returns:
            List[Tuple[InterfaceSpecific, str]]: A list of tuples with `InterfaceSpecific` instances and remaining
                                                designations.
        """
        if not full_designation.startswith('_'):
            return []

        full_designation = full_designation[1:]

        compos: list[tuple[InterfaceSpecific, str]] = []
        for implementation in all_implementations(InterfaceSpecific):
            try:
                compos.extend(implementation.from_designation(full_designation, components, sizes))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation}')
        return compos

    @staticmethod
    @abstractmethod
    def from_designation(
        full_designation: str, components: Components, sizes: Size
    ) -> list[tuple['InterfaceSpecific', str]]:
        """Parses the designation into specific interface format instances.

        Args:
            full_designation (str): The full designation string.
            components (Components): The pixel format components.
            sizes (Size): The sizes of the components.

        Returns:
            List[Tuple[InterfaceSpecific, str]]: A list of tuples with parsed instances and remaining designations.
        """


class Planar(InterfaceSpecific):
    """Represents the planar interface-specific format."""

    def designation(self) -> str:
        """Returns the planar designation."""
        return 'Planar'

    @staticmethod
    def from_designation(
        full_designation: str, components: Components, sizes: Size
    ) -> list[tuple['InterfaceSpecific', str]]:
        """Parses planar-specific designations.

        Args:
            full_designation (str): The full designation string.
            components (Components): The pixel format components.
            sizes (Size): The sizes of the components.

        Returns:
            List[Tuple[InterfaceSpecific, str]]: A list with a `Planar` instance if matched.
        """
        designation = 'Planar'
        if full_designation.startswith(designation):
            return [(Planar(), full_designation[len(designation) :])]
        return []

    def __eq__(self, other) -> bool:
        """Checks equality with another object."""
        return isinstance(other, Planar)


class SemiPlanar(InterfaceSpecific):
    """Represents the semi-planar interface-specific format."""

    def designation(self) -> str:
        """Returns the semi-planar designation."""
        return 'Semiplanar'

    @staticmethod
    def from_designation(
        full_designation: str, components: Components, sizes: Size
    ) -> list[tuple['InterfaceSpecific', str]]:
        """Parses semi-planar-specific designations.

        Args:
            full_designation (str): The full designation string.
            components (Components): The pixel format components.
            sizes (Size): The sizes of the components.

        Returns:
            List[Tuple[InterfaceSpecific, str]]: A list with a `SemiPlanar` instance if matched.
        """
        designation = 'Semiplanar'
        if full_designation.startswith(designation):
            return [(Planar(), full_designation[len(designation) :])]
        return []

    def __eq__(self, other) -> bool:
        """Checks equality with another object."""
        return isinstance(other, SemiPlanar)


class ComponentsSequencing(InterfaceSpecific):
    """Represents an interface-specific format based on component sequencing."""

    def __init__(self, components: tuple[str, ...]) -> None:
        """Initializes a `ComponentsSequencing` instance.

        Args:
            components (Tuple[str, ...]): A tuple of component names.
        """
        self.components: tuple[str, ...] = components

    def designation(self) -> str:
        """Returns the component sequencing designation."""
        return ''.join(self.components)

    @staticmethod
    def from_designation(
        full_designation: str, components: Components, sizes: Size
    ) -> list[tuple['InterfaceSpecific', str]]:
        """Parses component sequencing-specific designations.

        Args:
            full_designation (str): The full designation string.
            components (Components): The pixel format components.
            sizes (Size): The sizes of the components.

        Returns:
            List[Tuple[InterfaceSpecific, str]]: A list of parsed instances and remaining designations.
        """
        component_names: list[str] = [channel_name(components[i]) for i in range(len(components))]
        if not len(component_names):
            return []

        component_counts: dict[int, int] = {}
        extracted_components: list[str] = []

        i_f: int = 0
        while i_f <= len(full_designation):
            substr = full_designation[i_f:]

            matching_component_index: int
            matching_component_name: str
            for matching_component_index, matching_component_name in enumerate(component_names):
                if substr.startswith(matching_component_name):
                    break
            else:
                break

            extracted_components.append(matching_component_name)
            component_counts.setdefault(matching_component_index, 0)
            component_counts[matching_component_index] += 1

            i_f += len(matching_component_name)

        if len(component_counts) != len(sizes):
            return []

        if isinstance(components, GenicamComponents) and isinstance(components.location, LMNLocation):
            int_occurrences = components.location.fields
            if len(int_occurrences) != len(sizes):
                return []

            max_occurrence: int = max(int_occurrences)
            max_component_counts: int = max(component_counts.values())

            for i_c, int_occurrence in enumerate(int_occurrences):
                o_adjusted = int_occurrence * max_component_counts
                count_adjusted = component_counts[i_c] * max_occurrence

                if o_adjusted != count_adjusted:
                    return []

        return [(ComponentsSequencing(tuple(extracted_components)), full_designation[i_f:])]

    def __eq__(self, other) -> bool:
        """Checks equality with another object."""
        return isinstance(other, ComponentsSequencing) and self.components == other.components
