import traceback
import weakref
import re
import itertools
import time
import copy

import numpy as np
from builtins import AttributeError
from dataclasses import dataclass

import genicam.gentl
from genicam.gentl import (
    TimeoutException,
    NotAvailableException,
    ErrorException,
    ResourceInUseException,
    ClosedException,
    BusyException,
    AccessDeniedException,
    InvalidBufferException,
    InvalidHandleException,
    InvalidParameterException,
    GenericException,
    NotImplementedException,
    InvalidIdException,
    NoDataException,
    IoException,
    AbortException,
    LoadLibraryException,
    OutOfMemoryException,
    ResourceExhaustedException,
    InvalidValueException,
    ParsingChunkDataException,
    BufferTooSmallException,
    InvalidIndexException,
    InvalidAddressException,
    NotInitializedException,
)
from genicam.genapi import AccessException, InvalidArgumentException, OutOfRangeException, LogicalErrorException
from harvesters.core import DeviceInfo, ImageAcquirer
from logging import getLogger

from dataclasses import fields
from typing import Optional, Union
from functools import singledispatchmethod

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.sensor.data.camera_model import CameraModel
from dataspree.inference.sensor.data.pixel_format import PixelFormat
from dataspree.inference.sensor.driver import CameraDriver
from dataspree.inference.sensor.driver.harvesters_driver.multi_harvester import MultiHarvester, HarvesterLock
from dataspree.inference.sensor.data.sensor import SensorState
from dataspree.inference.sensor.data.configuration import (
    NotSupported,
    TemporarilyNotAvailable,
    NotAvailable,
    is_available,
    NotRetrieved,
    NotAvailableRead,
    NotAvailableWrite,
)
from dataspree.inference.sensor.data.camera import (
    Camera,
    TriggerMode,
    ContinuousAcquisitionMode,
    AcquisitionMode,
    CameraData,
    TriggerAcquisitionMode,
    SoftwareTrigger,
    HardwareTrigger,
    TriggerActivation,
    ManualGain,
    AutoGain,
    Gain,
    Exposure,
    AutoExposure,
    TimedExposure,
    Roi,
    ManualWhiteBalance,
    WhiteBalance,
    AutoWhiteBalance,
    Binning,
    ContrastAdjustments,
    TriggeredExposure,
    Brightness,
    ColorAdjustment,
    ColorAdjustmentConfiguration,
    TemperatureState,
    ManualBlackLevel,
    BlackLevel,
    AcquisitionStatus,
    FrameBurstStatus,
    ExposureStatus,
)
from dataspree.inference.sensor.data.exceptions import (
    IncompleteBufferError,
    SensorFeaturePermanentlyNotAvailable,
    NoFreeRois,
    SensorFeatureTemporarilyNotAvailable,
    SensorLost,
    DriverFeatureImplementationError,
    SensorNotFound,
    DriverFeatureNotImplemented,
    FeatureUnavailable,
    DriverInvalidState,
    SensorInvalidState,
    DriverStateInconsistent,
    InputException,
    ConfigurationError,
    SensorResourceTemporarilyNotAvailable,
    ConfigurationInconsistent,
    ResourceError,
    InvalidState,
    SensorResourcePermanentlyNotAvailable,
)
from dataspree.inference.plugins.harvester_utils import dump_genicam_settings, load_genicam_settings

logger = getLogger(__name__)


@dataclass
class Component:
    data: np.ndarray

    width: Optional[int]
    height: Optional[int]
    delivered_image_width: Optional[int]
    delivered_image_height: Optional[int]

    x_offset: Optional[int]
    y_offset: Optional[int]

    x_padding: Optional[int]
    y_padding: Optional[int]

    num_components_per_pixel: Optional[int]
    data_offset: Optional[int]

    data_format_value: Optional[str]
    data_format: Optional[str]

    def __post_init__(self):
        self.data = self.data.copy()


@dataclass
class ImageContent:
    # Is in ms; at least for the keyence cameras I am currently using it with.
    timestamp: Optional[int]
    timestamp_ns: Optional[int]
    components: list[Component]

    @property
    def timestamp_ms(self) -> Optional[int]:
        if self.timestamp_ns is not None:
            return self.timestamp_ns // int(1e6)

        return self.timestamp


def handle_harvesters_error(error: Union[GenericException, AttributeError], operation: str) -> None:
    """Transform a Harvester GenICam error into a suitable dataspree inference sensor error.

    Args:
        error (Exception): The Harvester error to be translated into a common inference error.
        operation (str): The operation that failed.

    Raises:
        ConfigurationError: If supplied values cannot be applied.
        InvalidState: If the sensor or driver is in an invalid state.
        ResourceError: If the resource is permanently or temporarily unavailable.
        FeatureUnavailable: If the sensor does not support the feature or cannot respond to it.
    """
    # if is_application_debug():
    #    traceback.print_exc()

    def msg(a: str = '') -> str:
        return f'Error {operation}: {type(error).__name__} {error}{a}'

    # feature temporarily not available

    if isinstance(error, (InvalidBufferException, InvalidHandleException)):
        raise SensorFeatureTemporarilyNotAvailable(msg()) from error

    # Resource temporarily not available

    if isinstance(error, (TimeoutException, ClosedException)):
        raise SensorResourceTemporarilyNotAvailable(msg()) from None

    if isinstance(error, (ResourceInUseException, OutOfMemoryException, BusyException)):
        raise SensorResourceTemporarilyNotAvailable(msg(' (resource in use or insufficient memory)')) from None

    if isinstance(error, AbortException):
        raise SensorResourceTemporarilyNotAvailable(msg()) from None

    if isinstance(error, ResourceExhaustedException):
        raise SensorResourceTemporarilyNotAvailable(msg()) from None

    if isinstance(error, NoDataException):
        raise SensorResourceTemporarilyNotAvailable(msg()) from None

    # feature Permanently not available

    if isinstance(error, AttributeError):
        raise SensorFeaturePermanentlyNotAvailable(msg()) from error
    if isinstance(error, (NotAvailableException, NotImplementedException)):
        raise SensorFeaturePermanentlyNotAvailable(msg()) from None

    # Configuration errors

    if isinstance(
        error, (InvalidArgumentException, InvalidParameterException, InvalidParameterException, InvalidValueException)
    ):
        raise ConfigurationError(msg()) from None

    # Invalid state

    if isinstance(error, (AccessException, AccessDeniedException)):
        raise SensorInvalidState(msg()) from None

    if isinstance(error, (InvalidHandleException, InvalidIdException, InvalidAddressException, InvalidIndexException)):
        # Raised when an invalid handle is passed to a function or when a handle is no longer valid
        # (e.g., if the connection to a device is lost).
        raise SensorNotFound(msg()) from None

    if isinstance(error, NotInitializedException):
        raise SensorResourceTemporarilyNotAvailable(msg(' (sensor not initialized or invalid handle)')) from None

    if isinstance(error, ParsingChunkDataException):
        raise SensorResourceTemporarilyNotAvailable(msg()) from None

    if isinstance(error, BufferTooSmallException):
        raise ConfigurationInconsistent(msg()) from None

    if isinstance(error, LoadLibraryException):
        raise SensorInvalidState(msg()) from None

    if isinstance(error, IoException):
        raise SensorResourceTemporarilyNotAvailable(msg()) from None

    # Fallback

    if isinstance(error, (GenericException, ErrorException)):
        raise SensorFeatureTemporarilyNotAvailable(msg(' (interface temporarily unavailable)')) from None

    raise DriverFeatureImplementationError(msg(' (unhandled error)')) from None


class HarvestersDriver(CameraDriver):
    """Access to this driver is must occur from a single thread.

    All the set_ methods receive the value that is supposed to be applied as a an

    """

    def __init__(self, cti_file: str) -> None:
        CameraDriver.__init__(self)
        self.cti_file: str = cti_file
        self._harvester_registry: Optional[tuple[weakref.ReferenceType[HarvesterLock], int]] = None
        self._image_acquirers: dict[str, ImageAcquirer] = {}

    #
    # General functionality
    #

    def setup(self) -> 'HarvestersDriver':
        """Initialize the driver. This function must be called before any other functionality can be used.

        Initializes the list of sensors and driver internals.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.
        """
        if self._harvester_registry is not None:
            raise DriverInvalidState('Harvester registry not destroyed. Unable to set up.')

        harvester_registry = MultiHarvester.registry.setup()
        try:
            self._harvester_registry = (harvester_registry, harvester_registry.register(cti_files={self.cti_file}))

        except DriverInvalidState as e:
            self._harvester_registry = None
            raise e from None

        except Exception as e:
            self._harvester_registry = None
            raise DriverFeatureImplementationError(f'An exception occurred during harvester registration: {e}') from e

        else:
            self.create_sensors()
            return self

    def destroy(self) -> None:
        """Destroys the driver. After calling destroy, no other function can be called successfully aside from setup.

        Destroys the list of sensors and unregisters driver internals.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.
        """
        if self._harvester_registry is None:
            raise DriverInvalidState('The harvester driver is not set up.')

        try:
            self._harvester_registry[0].deregister(self._harvester_registry[1])

        except ReferenceError:
            pass

        except InputException as e:
            logger.error(f'An error occurred while attempting to unregister cti file: {type(e).__name__} {e}.')

        except (AttributeError, GenericException) as e:
            logger.fatal(
                f'An unknown error occurred while attempting to unregister cti file:' f' {type(e).__name__} {e}.'
            )

        finally:
            try:
                MultiHarvester.registry.destroy()
            except InputException as e:
                logger.fatal(
                    f'An input exception occurred while attempting to destroy the '
                    f'registry: {type(e).__name__}: {e}.'
                )
            except Exception as e:
                logger.fatal(
                    f'An error occurred while attempting to destroy the registry: '
                    f'{type(e).__name__}: {e}.'
                    f'Are you using the stock version of harvesters? '
                    f'Use the patched version that is provided in the requirements. '
                    f'Otherwise you will probably never find a sensor again in this session.'
                )
            finally:
                self._harvester_registry = None
                self.destroy_sensors()

    def retrieve_sensors(self) -> None:
        """Retrieve all available sensors and update self.sensors.

        Does not refresh sensor attributes.

        Raises:
            DriverInvalidState: if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.

            ResourceError: if the resource is currently not available.
        """
        if self._harvester_registry is None:
            raise DriverInvalidState('The harvester driver is not set up.')

        device_info_list: list[DeviceInfo]
        with self._harvester_registry[0] as unique_lock:
            try:
                unique_lock.harvester.update()

            except (AttributeError, GenericException) as e:
                try:
                    handle_harvesters_error(e, 'update unique lock')

                except (DriverInvalidState, DriverFeatureImplementationError, ResourceError) as ex:
                    raise ex from e

                except InputException as ex:
                    raise DriverFeatureImplementationError('unexpected input error') from ex

            # @dev: this cannot be serialized / copied
            device_info_list = unique_lock.harvester.device_info_list
            discovered_ids: list[str] = [
                info.property_dict.get('serial_number', info.property_dict.get('id_')) for info in device_info_list
            ]

            for sensor_id in self.sensors:
                if sensor_id not in discovered_ids:
                    self.sensors[sensor_id].state = SensorState.UNAVAILABLE

            for internal_device_id, info in enumerate(device_info_list):
                device_properties = {}
                for k, v in info.property_dict.items():
                    try:
                        device_properties[k] = copy.deepcopy(v)
                    except TypeError:
                        pass

                identifier = info.property_dict.get('serial_number', info.property_dict.get('id_'))
                try:
                    state = self.sensors[identifier].state
                    if state == SensorState.UNAVAILABLE:
                        state = SensorState.DISCOVERED
                except KeyError:
                    state = SensorState.DISCOVERED

                camera = Camera(
                    identifier=identifier,
                    internal_device_id=internal_device_id,
                    name=info.property_dict.get('display_name'),
                    manufacturer=info.property_dict.get('vendor'),
                    model=info.property_dict.get('model'),
                    firmware_version=info.property_dict.get('version'),
                    state=state,
                    device_properties=device_properties,
                )

                try:
                    # Update the fetched camera properties while retaining old properties, that were not retrieved.
                    if camera.identifier in self.sensors:
                        assert self.sensors[camera.identifier].identifier == camera.identifier
                        self.sensors[camera.identifier].name = camera.name
                        self.sensors[camera.identifier].internal_device_id = internal_device_id
                        self.sensors[camera.identifier].manufacturer = (camera.manufacturer,)
                        self.sensors[camera.identifier].model = (camera.model,)
                        self.sensors[camera.identifier].firmware_version = (camera.firmware_version,)
                        self.sensors[camera.identifier].device_properties.update(camera.device_properties)
                        self.sensors[camera.identifier].state = camera.state

                    else:
                        self.sensors.add_sensor(camera)
                except InputException as e:
                    logger.error(f'Unable to add sensor {camera}: {type(e).__name__} {e}')
                    if is_application_debug():
                        traceback.print_exc()

    def reset_driver(self) -> None:
        """Reset the driver.

        Temporarily disconnects and stops all cameras. This function can be called at any point in time.

        1. Driver internals are reset
        2. The driver stops and disconnects all sensors.
        3. Call to `destroy`
        4. Call to `setup`
        5. Call to `retrieve_sensors`
        6. The driver reconnects to all previously connected sensors and starts all previously started sensors.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            SensorInvalidState if the sensor is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.

            ResourceError: if the resource is currently not available.
        """
        MultiHarvester.registry.reset()

        if self.sensors is not None:
            connected_sensors: dict[str, Camera] = {
                sid: s for sid, s in self.sensors.items() if s.state == SensorState.CONNECTED
            }
            started_sensors: dict[str, Camera] = {
                sid: s for sid, s in self.sensors.items() if s.state == SensorState.STARTED
            }

            for sensor in started_sensors.values():
                try:
                    self.stop(sensor)
                    self.disconnect(sensor)
                except InputException as e:
                    logger.error(
                        f'An exception occurred while attempting to stop/disconnect the sensor: {type(e).__name__}: '
                        f'{e}'
                    )
                    if is_application_debug():
                        traceback.print_exc()

            for sensor in connected_sensors.values():
                try:
                    self.disconnect(sensor)
                except InputException as e:
                    logger.error(
                        f'An exception occurred while attempting to disconnect the sensor: {type(e).__name__}: {e}'
                    )
                    if is_application_debug():
                        traceback.print_exc()

            try:
                self.destroy()
            except InputException as e:
                logger.error(f'An exception occurred while attempting to destroy the driver: {type(e).__name__}: {e}')
                if is_application_debug():
                    traceback.print_exc()

            self.setup()
            self.retrieve_sensors()

            sensor_errors: dict[str, InputException] = {}
            for sid in connected_sensors.keys():
                try:
                    try:
                        self.connect(self.sensors[sid])
                    except KeyError:
                        raise SensorNotFound
                except InputException as e:
                    sensor_errors[sid] = e

            for sid in started_sensors:
                try:
                    try:
                        self.connect(self.sensors[sid])
                        self.start(self.sensors[sid])
                    except KeyError:
                        raise SensorNotFound
                except InputException as e:
                    sensor_errors[sid] = e

            if len(sensor_errors):
                err: str = ', '.join(f'"{sn}": [{type(err).__name__}] {err}' for sn, err in sensor_errors.items())
                raise SensorFeatureTemporarilyNotAvailable(f'Could not initialize sensors: {err}')

    def _node_map(self, sensor: Camera, ensure_state: Optional[set[SensorState]] = None):
        """Retrieve the sensor's node map.

        Arguments:
            sensor (Camera): the sensor that belongs to node map that is supposed to be returned.

            ensure_state: Optional[Set[SensorState]] a set of states the camera state should belong to.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not in the disjunction of `ensure_state` and
                `{STARTED, CONNECTED}`.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        try:
            ia = self._image_acquirer(sensor, ensure_state)
            return ia.remote_device.node_map
        except AttributeError:
            raise SensorNotFound from None
        except GenericException as e:
            raise SensorNotFound from e

    def _image_acquirer(self, sensor: Camera, ensure_state: Optional[set[SensorState]] = None) -> ImageAcquirer:
        """Retrieve the image acquirer that belongs to a specific sensor.

        Arguments:
            sensor (Camera): the sensor that belongs to the image acquirer that is supposed to be returned.

            ensure_state: Optional[Set[SensorState]] a set of states the camera state should belong to.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not in the disjunction of `ensure_state` and
                `{STARTED, CONNECTED}`.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)

        accepted_states = {
            a for a in {SensorState.STARTED, SensorState.CONNECTED} if ensure_state is None or a in ensure_state
        }
        if sensor.state not in accepted_states:
            raise SensorInvalidState.from_sensor(sensor, accepted_states)

        optional_ia: Optional[ImageAcquirer] = self._image_acquirers.get(sensor.identifier)
        if optional_ia is None:
            raise DriverStateInconsistent

        return optional_ia

    #
    # Sensor control
    #

    def connect(self, sensor: Camera) -> None:
        """Connect to the provided sensor.

        This creates a new image acquirer and adds it to the mapping of all image acquirers.
        After the connection is established, the sensor's attributes are fetched and stored in the sensor instance.

        Arguments:
            sensor (Camera) The sensor which is supposed to be connected.

        Raises:
            InvalidState If the sensor or driver is in an invalid state


            DriverFeatureImplementationError: If an exception occurred in the implementation.

            ResourceError: If the requested resource is not available.

        See Also:
            `self.refresh_attributes`: Function used for gathering the attributes from the remote sensor.
        """
        sensor = self._sensor(sensor)

        if sensor.state != SensorState.DISCOVERED:
            raise SensorInvalidState.from_sensor(sensor, {SensorState.DISCOVERED})

        with self._harvester_registry[0] as unique_lock:
            try:
                for i in itertools.count(start=0):
                    try:
                        if not is_available(sensor.internal_device_id):
                            raise SensorLost(sensor)

                        self._image_acquirers[sensor.identifier] = unique_lock.harvester.create(
                            sensor.internal_device_id
                        )
                        break
                    except ClosedException as e:
                        if i == 4:
                            raise e from None

            except (GenericException, AttributeError) as e:
                try:
                    handle_harvesters_error(e, 'connect')
                except (InvalidState, DriverFeatureImplementationError, ResourceError) as ex:
                    raise ex from None

                except InvalidState as ex:
                    raise DriverFeatureImplementationError(f'Unexpected exception: {ex}')

            # TODO: If i set the value before, refresh fails!
            # self.refresh_attributes(sensor)
            sensor.state = SensorState.CONNECTED
            self.refresh_attributes(sensor)

        # self.activate_chunks(sensor)

    def disconnect(self, sensor: Camera) -> None:
        """Disconnect from the provided sensor.

        This removes the image acquirer from the mapping of all image acquirers.

        Arguments:
            sensor (Camera) The sensor which is supposed to be disconnected.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            DriverFeatureImplementationError: If an exception occurred in the implementation.

            ResourceError: If the requested resource is not available.
        """
        sensor = self._sensor(sensor)

        if sensor.state != SensorState.CONNECTED:
            raise SensorInvalidState.from_sensor(sensor, {SensorState.CONNECTED})

        image_acquirer = self._image_acquirers.pop(sensor.identifier)
        try:
            image_acquirer.destroy()
            sensor.state = SensorState.DISCOVERED
        except (AttributeError, GenericException):
            sensor.state = SensorState.DISCOVERED
            raise SensorLost(sensor)

    def start(self, sensor: Camera) -> None:
        """Start the data acquisition from the provided sensor.

        Arguments:
            sensor (Camera): the sensor that is supposed to be started.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            FeatureUnavailable: If the feature is not available for the sensor or the driver.

            ResourceError: If the requested resource is not available.
        """
        sensor = self._sensor(sensor)
        image_acquirer = self._image_acquirer(sensor, {SensorState.CONNECTED})
        try:
            image_acquirer.start()
            self._node_map(sensor, {SensorState.CONNECTED}).AcquisitionStart.execute()
        except (AttributeError, GenericException) as e:
            try:
                image_acquirer.stop()
            except (AttributeError, GenericException):
                pass

            try:
                handle_harvesters_error(e, 'start')
            except (InvalidState, DriverFeatureImplementationError, ResourceError) as ex:
                raise ex from None

            except InvalidState as ex:
                raise DriverFeatureImplementationError(f'Unexpected exception: {ex}')

        try:
            sensor.state = SensorState.STARTED
        except InputException as e:
            raise DriverFeatureImplementationError() from e

    def stop(self, sensor: Camera) -> None:
        """Stop the data acquisition from the provided sensor.

        Arguments:
            sensor (Camera): the sensor that is supposed to be stopped.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            FeatureUnavailable: If the feature is not available for the sensor or the driver.

            ResourceError: If the requested resource is not available.
        """
        sensor = self._sensor(sensor)
        image_acquirer = self._image_acquirer(sensor, {SensorState.STARTED})
        try:
            self._node_map(sensor, {SensorState.STARTED}).AcquisitionStop.execute()
            image_acquirer.stop()

        except (AttributeError, GenericException, InputException):
            sensor.state = SensorState.CONNECTED
            raise SensorLost(sensor)

        try:
            sensor.state = SensorState.CONNECTED
        except InputException as e:
            raise DriverFeatureImplementationError() from e

    def reset(self, sensor: Camera) -> None:
        """Reset the provided sensor.

        A reset
        1. stops image acquisition if image acquisition was started,
        2. performs a sensor reset,
        3. disconnects form the sensor,
        4. connects to the sensor, and
        5. optionally restarts the sensor if image acquisition was started.

        Arguments:
            sensor (Camera): the sensor that is supposed to be reset.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """
        sensor = self._sensor(sensor)
        was_started: bool = sensor.state == SensorState.STARTED

        node_map = self._node_map(sensor)
        if was_started:
            self.stop(sensor)

        try:
            node_map.DeviceReset.execute()
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'execute device reset')
            except (InvalidState, ResourceError, FeatureUnavailable) as ex:
                raise ex from None
            except InputException as ex:
                raise DriverFeatureImplementationError(
                    f'An unexpected exception occurred: ' f'{type(ex).__name__}: {ex}'
                ) from ex

        self.disconnect(sensor)
        self.connect(sensor)
        if was_started:
            self.start(sensor)

    #
    # Sensor properties
    #

    def refresh_attributes(self, sensor: Camera) -> None:
        """Refresh all available attributes.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        CameraDriver.refresh_attributes(self, sensor)
        try:
            self.refresh_camera_model(sensor)
        except SensorNotFound:
            pass

    def refresh_camera_model(self, sensor: Camera):
        """Updates the camera model with values retrieved from the remote instance."""
        sensor = self._sensor(sensor)
        self.refresh_available_pixel_formats(sensor)
        self.refresh_available_configuration_names(sensor)
        self.refresh_available_acquisition_burst_modes(sensor)
        self.refresh_available_acquisition_burst_modes(sensor)
        self.refresh_available_gpio_lines(sensor)
        # TODO: Add all options here.

    def refresh_available_pixel_formats(self, sensor: Camera) -> None:
        """Refresh the pixel formats that are available for the camera model.

        Arguments:
            sensor (Camera): the sensor for which the camera model is to be updated.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)

        try:
            node_map = self._node_map(sensor)
        except SensorInvalidState:
            raise SensorNotFound

        try:
            allowed_formats: tuple[str, ...] = node_map.PixelFormat.symbolics
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'get pixel format')
            except (
                SensorResourceTemporarilyNotAvailable,
                SensorFeatureTemporarilyNotAvailable,
                SensorFeaturePermanentlyNotAvailable,
                SensorResourcePermanentlyNotAvailable,
            ):
                return
            except InputException:
                raise SensorLost(sensor)
            else:
                raise SensorLost(sensor)

        if not len(allowed_formats):
            return

        configured_pixel_formats: dict[str, PixelFormat] = {
            f.designation(): f for f in sensor.camera_model.available_pixel_formats
        }
        if sorted(configured_pixel_formats.keys()) != sorted(allowed_formats):
            prev_pixel_fromats = ', '.join(configured_pixel_formats.keys())
            logger.info(
                f'Updating the registered pixel formats for the camera model ({prev_pixel_fromats}) '
                f'to the retrieved pixel formats ({", ".join(allowed_formats)}).'
            )

            try:
                available_pixel_formats: list[PixelFormat] = [PixelFormat.from_name(f) for f in allowed_formats]
            except DriverFeatureImplementationError:
                return

            sensor.camera_model.available_pixel_formats = available_pixel_formats

    def refresh_available_configuration_names(self, sensor: Camera) -> None:
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        config_names: Optional[list[str]]
        try:
            config_names = list(node_map.UserSetSelector.symbolics)
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'get configuration names')
            except (
                SensorResourceTemporarilyNotAvailable,
                SensorFeatureTemporarilyNotAvailable,
                SensorFeaturePermanentlyNotAvailable,
                SensorResourcePermanentlyNotAvailable,
            ):
                return
            except InputException:
                raise SensorLost(sensor)
            else:
                raise SensorLost(sensor)

        if config_names is not None:
            available_configuration_names = config_names

            available_configuration_names_writable = [
                f for f in sensor.camera_model.available_configuration_names_writable if f in config_names
            ]

            available_configuration_names_read_only = [
                f for f in sensor.camera_model.available_configuration_names_read_only if f in config_names
            ]

            if sorted(sensor.camera_model.available_configuration_names) != sorted(available_configuration_names):
                logger.info(f'Updated configuration names to {", ".join(available_configuration_names)}.')
                sensor.camera_model.available_configuration_names = available_configuration_names

            if sorted(sensor.camera_model.available_configuration_names_writable) != sorted(
                available_configuration_names_writable
            ):
                logger.info(
                    f'Updated writable configuration names to {", ".join(available_configuration_names_writable)}.'
                )
                sensor.camera_model.available_configuration_names_writable = available_configuration_names_writable

            if sorted(sensor.camera_model.available_configuration_names_read_only) != sorted(
                available_configuration_names_read_only
            ):
                logger.info(
                    f'Updated read-only configuration names to {", ".join(available_configuration_names_read_only)}.'
                )
                sensor.camera_model.available_configuration_names_read_only = available_configuration_names_read_only

    def refresh_available_acquisition_burst_modes(self, sensor: Camera) -> None:
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        burst_modes: Optional[list[str]]
        try:
            burst_modes = list(node_map.BslAcquisitionBurstMode.symbolics)
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'get acquistion burst modes')
            except (
                SensorResourceTemporarilyNotAvailable,
                SensorFeatureTemporarilyNotAvailable,
                SensorFeaturePermanentlyNotAvailable,
                SensorResourcePermanentlyNotAvailable,
            ):
                return
            except InputException:
                raise SensorLost(sensor)
            else:
                raise SensorLost(sensor)

        sensor.camera_model.acquisition_burst_modes = burst_modes

    def refresh_available_gpio_lines(self, sensor: Camera):
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        symbolics: Optional[list[str]]
        try:
            symbolics = list(node_map.LineSelector.symbolics)
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'get gpio lines')
            except (
                SensorResourceTemporarilyNotAvailable,
                SensorFeatureTemporarilyNotAvailable,
                SensorFeaturePermanentlyNotAvailable,
                SensorResourcePermanentlyNotAvailable,
            ):
                return
            except InputException:
                raise SensorLost(sensor)
            else:
                raise SensorLost(sensor)

        if symbolics is not None and sorted(symbolics) != sorted(sensor.camera_model.gpio_lines):
            logger.info(f'Updated read-only configuration names to {", ".join(symbolics)}.')
            sensor.camera_model.gpio_lines = symbolics

    def activate_chunks(self, sensor: Camera, activate: bool = True) -> None:
        """Activate chunks for transmission of metainformation.

        Arguments:
            sensor (Camera): The sensor for which chunks are supposed to be activated.

            activate (bool):  Whether to activate or deactivate the transmission of chunks.
        """
        try:
            sensor = self._sensor(sensor)
            camera = self._node_map(sensor)
            camera.ChunkModeActive.value = activate
        except InputException:
            raise SensorLost(sensor)

        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'activating chunks')
            except (SensorResourceTemporarilyNotAvailable, SensorFeatureTemporarilyNotAvailable):
                return
            except InputException:
                raise SensorLost(sensor)
            else:
                raise SensorLost(sensor)

        else:
            for value in camera.ChunkSelector.symbolics:
                try:
                    camera.ChunkSelector.value = value
                    camera.ChunkEnable.value = activate
                except (AttributeError, AccessException):
                    ...

                except GenericException:
                    if is_application_debug():
                        traceback.print_exc()

    #
    # Statistics values
    #

    @staticmethod
    def retrieve_effective_framerate(node_map) -> Union[float, NotAvailableRead]:
        """The Resulting Acquisition Frame Rate camera feature allows you to determine the estimated
        frame rate with the current camera settings.

        This is useful, for example, if you want to know how long you have to wait between triggers.

        The frame rate is expressed in hertz (Hz).

        Factors that limit the acquisition frame rate can be found int in the documentation.

        Arguments:
            node_map: node map of the sensor

        References:
            [1] Resulting Acquisition Frame Rate. (2024)
                Retrieved from https://docs.baslerweb.com/resulting-acquisition-frame-rate
        """
        try:
            return node_map.ResultingFrameRate.value
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'retrieve_effective_framerate')
            except (SensorFeatureTemporarilyNotAvailable, SensorResourceTemporarilyNotAvailable):
                return TemporarilyNotAvailable()
            except InputException:
                return NotSupported()

    @staticmethod
    def retrieve_effective_linerate(node_map) -> Union[float, NotAvailableRead]:
        """Return a camera's acquisition rate line rate.

        The Resulting Acquisition Line Rate camera feature allows you to determine the estimated line rate with the
        current camera settings.

        This is useful, for example, if you want to know how long you have to wait between triggers.

        The line rate is expressed in hertz (Hz).

        Factors that limit the acquisition line rate can be found int in the documentation.

        References:
            [1] Resulting Acquisition Line Rate. (2024)
                Retrieved from https://docs.baslerweb.com/resulting-acquisition-line-rate

        Arguments:
            node_map: node map of the sensor

        Returns:
            Union[float, NotAvailableRead]: the camera's acquisition rate.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        try:
            return node_map.BslResultingAcquisitionLineRate.value
        except (AttributeError, GenericException) as e:
            try:
                handle_harvesters_error(e, 'retrieve_effective_linerate')
            except (SensorFeatureTemporarilyNotAvailable, SensorResourceTemporarilyNotAvailable):
                return TemporarilyNotAvailable()
            except InputException:
                return NotSupported()

    def fetch_resulting_effective_acquisition_rate(self, sensor: Camera) -> Union[float, NotAvailableRead]:
        """Return a camera's acquisition rate.

        Arguments:
            sensor (T): The sensor from which we gather the acquisition rate.

        Returns:
            Union[float, NotAvailableRead]: the camera's acquisition rate.

        Raises:
            SensorNotFound if the sensor is currently not found.

        See:
            `cls.retrieve_effective_framerate`: Function used for fetching the framerate.

            `cls.retrieve_effective_linerate`: Function used for fetching the linerate.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        effective_framerate = self.retrieve_effective_framerate(node_map)
        return effective_framerate if is_available(effective_framerate) else self.retrieve_effective_linerate(node_map)

    #
    # Acquisition Mode
    #

    def fetch_acquisition_mode(self, sensor: Camera) -> Union[AcquisitionMode, NotAvailable]:
        """Retrieve the currently active acquisition mode from the specified sensor.

        Args:
            sensor (Camera): The sensor object from which to retrieve the acquisition mode.

        Returns:
            AcquisitionMode: The current acquisition mode of the sensor.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound: If the specified sensor is not found.
            SensorInvalidState: If the sensor's state is neither `CONNECTED` nor `STARTED`.
            DriverStateInconsistent: If the sensor connection objects are missing despite the state indicating their
                presence.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        # max rate
        max_rate: Optional[Union[float, NotAvailable]] = self.fetch_acquisition_max_rate(node_map)

        # Determine the configuration of all currently available triggers
        trigger_modes: dict[Optional[str], Optional[TriggerMode]] = {None: None}
        camera_model = sensor.camera_model
        for trigger_name in filter(lambda a: a, camera_model.acquisition_trigger_names):
            trigger_information = self.fetch_trigger(node_map, sensor.camera_model, trigger_name)
            if is_available(trigger_information) and trigger_information[1]:
                trigger_modes[trigger_name] = trigger_information[0]
            else:
                trigger_modes[trigger_name] = None

        try:
            acquisition_mode = node_map.AcquisitionMode.value
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

        # TODO: Investigate, if we can set specific triggers also in continuous mode?
        #       I would hope / suppose that we would receive an Access Error
        #       If this is changed, check _apply_continuous_acquisition
        if acquisition_mode == 'Continuous' and all(map(lambda a: not is_available(a), trigger_modes.values())):
            return ContinuousAcquisitionMode(max_rate)

        # TODO: Investigate what happens if FrameActive and FrameStart is set simultaneously
        #       with a device that supports both FrameActive and FrameStart.
        # TODO: Are there devices that don't support FrameStart?

        if (camera_model.frame_start_trigger is not None) and (camera_model.line_start_trigger is not None):
            logger.error('Both frame start trigger and line start trigger are available.')

        try:
            trigger_frame_start, trigger_frame_end = self.determine_trigger_modes(
                trigger_modes[camera_model.frame_start_trigger or camera_model.line_start_trigger],
                trigger_modes[camera_model.frame_stop_trigger],
                trigger_modes[camera_model.frame_active_trigger],
            )

            trigger_frame_burst_start, trigger_frame_burst_end = self.determine_trigger_modes(
                trigger_modes[camera_model.frame_burst_start_trigger],
                trigger_modes[camera_model.frame_burst_stop_trigger],
                trigger_modes[camera_model.frame_burst_active_trigger],
            )

            if trigger_modes[camera_model.frame_start_trigger] is None:
                trigger_frame_start = SoftwareTrigger(camera_model.software_trigger_frame_start, NotSupported())

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        except SensorFeatureTemporarilyNotAvailable:
            return TemporarilyNotAvailable()

        except FeatureUnavailable as e:
            logger.error(f'Could not determine trigger start / stop timestamp: {e}')
            return NotSupported()

        burst_frame_count: Union[int, NotAvailable]
        try:
            burst_frame_count = node_map.AcquisitionBurstFrameCount.value
        except AttributeError:
            burst_frame_count = NotSupported()
        except AccessException:
            burst_frame_count = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        burst_mode: Union[str, NotAvailable]
        try:
            burst_mode = node_map.BslAcquisitionBurstMode.value
        except AttributeError:
            burst_mode = NotSupported()
        except AccessException:
            burst_mode = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        return TriggerAcquisitionMode(
            max_rate=max_rate,
            burst_frame_count=burst_frame_count,
            burst_mode=burst_mode,
            trigger_start=trigger_frame_start,
            trigger_frame_end=trigger_frame_end,
            trigger_frame_burst_start=trigger_frame_burst_start,
            trigger_frame_burst_end=trigger_frame_burst_end,
        )

    @staticmethod
    def fetch_acquisition_max_framerate(node_map) -> Union[float, NotAvailable, None]:
        """Retrieve the maximum frame rate of the sensor associated with the given image acquirer.

        Args:
            node_map: The node map object associated with the sensor.

        Returns:
            Optional[float]: The maximum frame rate of the sensor if it is enabled and accessible;
                             otherwise, `None` if the frame rate cannot be determined or an error occurs.
        """
        try:
            if node_map.AcquisitionFrameRateEnable.value:
                return node_map.AcquisitionFrameRate.value
            else:
                return None
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

    @staticmethod
    def fetch_acquisition_max_linerate(node_map) -> Union[float, NotAvailable, None]:
        """Retrieve the maximum line rate of the sensor associated with the given image acquirer.

        Args:
            node_map: The image acquirer object associated with the sensor.

        Returns:
            Optional[float]: The maximum frame rate of the sensor if it is enabled and accessible;
                             otherwise, `None` if the frame rate cannot be determined or an error occurs.
        """
        try:
            if node_map.AcquisitionLineRateEnable.value:
                return node_map.AcquisitionLineRate.value
            else:
                return None

        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

    @classmethod
    def fetch_acquisition_max_rate(cls, node_map) -> Union[float, NotAvailable]:
        """Fetch the maximum acquisition rate (frame rate or line rate) for the given image acquirer.

        This method first attempts to retrieve the maximum frame rate of the sensor associated with
        the provided image acquirer. If the frame rate cannot be determined, it then attempts to
        retrieve the maximum line rate.

        Args:
            node_map : The node map associated with the sensor from which the maximum rate is to
                       be fetched.

        Returns:
            Optional[float]: The maximum acquisition rate of the sensor. This could be the frame rate
                             or the line rate depending on availability. Returns `None` if neither
                             the frame rate nor the line rate can be determined.

        See Also:
            `cls.fetch_acquisition_max_framerate` retrieves the max framerate
            `cls.fetch_acquisition_max_linerate  retrieves the max linerate

        Notes:
            - The returned rate is in Hz (e.g., frames per second or lines per second).
        """
        effective_framerate: Union[float, NotAvailable] = cls.fetch_acquisition_max_framerate(node_map)
        if is_available(effective_framerate):
            return effective_framerate
        return cls.fetch_acquisition_max_linerate(node_map)

    @staticmethod
    def fetch_trigger(node_map, camera_model: CameraModel, key: str) -> Union[tuple[TriggerMode, bool], NotAvailable]:
        """Retrieve the trigger mode configuration for the specified key from the given image acquirer.

        Args:
            node_map: The node map object used to access the sensor's
            trigger settings.
            camera_model (CameraModel): The camera model object used to determine trigger source types.
            key (str): The trigger selector key to set for retrieving the trigger mode.

        Returns:
            Optional[Tuple[TriggerMode, bool]]: An instance of `SoftwareTrigger` or `HardwareTrigger` and an
                indication on whether the trigger is active

        """
        try:
            node_map.TriggerSelector.value = key
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

        try:
            active: bool = HarvestersDriver.parse_trigger_mode(node_map.TriggerMode.value)
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except ValueError:
            return NotRetrieved()

        try:
            src: str = node_map.TriggerSource.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

        trigger_delay: Union[float, NotAvailable]
        try:
            trigger_delay = node_map.TriggerDelay.value
        except AttributeError:
            trigger_delay = NotSupported()
        except AccessException:
            trigger_delay = TemporarilyNotAvailable()

        if src in camera_model.software_trigger_sources_generic:
            return SoftwareTrigger(src, trigger_delay), active

        if src == camera_model.software_trigger_frame_start:
            return SoftwareTrigger(src, trigger_delay), active

        try:
            trigger_activation: str = node_map.RisingEdge.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

        return HardwareTrigger(
            src, trigger_delay, HarvestersDriver.parse_trigger_activation(trigger_activation)
        ), active

    @staticmethod
    def parse_trigger_mode(trigger_str: str) -> bool:
        """Parse a trigger mode string into a boolean value.

        This method converts a trigger mode string, into a corresponding boolean value.
        If the string is not recognized, a `ValueError` is raised.

        Args:
            trigger_str (str): The trigger mode string to parse.

        Returns:
            bool: `True` exactly if the string indicates that the trigger is enabled.

        Raises:
            DriverStateException: if the driver is in an invalid state

            ValueError: If the trigger string is not recognized
        """
        try:
            return {'On': True, 'Off': False}[trigger_str]
        except KeyError:
            raise ValueError(f'Unknown trigger string {trigger_str}.')

    @staticmethod
    def trigger_mapping() -> dict[str, TriggerActivation]:
        """Returns the mapping from trigger keys to TriggerActivation instances."""
        return {
            'RisingEdge': TriggerActivation.RISING_EDGE,
            'FallingEdge': TriggerActivation.FALLING_EDGE,
            'AnyEdge': TriggerActivation.ANY_EDGE,
            'LevelHigh': TriggerActivation.LEVEL_HIGH,
            'LevelLow': TriggerActivation.LEVEL_LOW,
        }

    @classmethod
    def parse_trigger_activation(cls, trigger_activation_str: str) -> TriggerActivation:
        """Parse a trigger activation string into a corresponding `TriggerActivation` enum value.

        Args:
            trigger_activation_str (str): The trigger activation string to parse. This should match one of the
                                          keys in the `trigger_mapping`.

        Returns:
            TriggerActivation: The corresponding `TriggerActivation` enum value for the provided string.

        Raises:
            DriverStateException: if the driver is in an invalid state

            ValueError: If the `trigger_activation_str` does not match any key in the trigger mapping.

        See:
            The mapping is retrieved from the `HarvestersDriver.trigger_mapping` method.
        """
        try:
            return cls.trigger_mapping()[trigger_activation_str]
        except KeyError:
            raise ValueError(f'Unknown trigger activation string {trigger_activation_str}.')

    @classmethod
    def serialize_trigger_activation(cls, trigger_activation: TriggerActivation) -> str:
        """Convert a `TriggerActivation` enum value into its corresponding string representation.

        Args:
            trigger_activation (TriggerActivation): The `TriggerActivation` enum value to convert to a string.

        Returns:
            str: The string representation of the provided `TriggerActivation` enum value.

        Raises:
            DriverStateException: if the driver is in an invalid state

            ValueError: If the `trigger_activation` does not match any value in the predefined mapping.

        See:
            The mapping is retrieved from the `HarvestersDriver.trigger_mapping` method.
        """
        try:
            return {v: k for k, v in cls.trigger_mapping().items()}[trigger_activation]
        except KeyError:
            raise ValueError(f'Trigger activation {trigger_activation} not implemented.')

    @staticmethod
    def determine_trigger_modes(
        trigger_start: Optional[TriggerMode], trigger_end: Optional[TriggerMode], trigger_active: Optional[TriggerMode]
    ) -> tuple[Optional[TriggerMode], Optional[TriggerMode]]:
        """Determine the start and end trigger modes based on the active trigger mode configuration.

        Args:
            trigger_start (Optional[TriggerMode]): The trigger for starting an operation. If `trigger_active` is not
                None, trigger_start will be determined from it.

            trigger_end (Optional[TriggerMode]): The trigger for stopping an operation. If `trigger_active` is not
                None, trigger_end will be determined from it.

            trigger_active (Optional[TriggerMode]): The trigger, that indicates the duration in which the trigger
                is active. We determine the trigger_start and trigger_end values from this.

        Returns:
            Tuple[Optional[TriggerMode], Optional[TriggerMode]]:
                A tuple containing the determined start and end trigger modes.

        Raises:
            DriverStateException: if the driver is in an invalid state

            DriverFeatureNotImplemented: If combinations are not implemented.

            SensorFeatureCurrentlyNotAvailable: If the sensor is currently running a mode that
                                                does not rely on hardware triggered acquisition
                                                even though trigger active is configured.

        """
        if trigger_active is not None:
            # TODO: Investigate, what happens if the trigger frame active is configured as any
            #       edge, level high or level low.
            # TODO: also check if rising edge and falling edge does what I think it does here.
            # TODO: We characterize software triggers as not having rising edge / falling edge
            #       so software triggers should not have a duration.
            if not isinstance(trigger_active, HardwareTrigger):
                raise SensorFeatureTemporarilyNotAvailable('Did not consider a software trigger to have a duration.')

            trigger_start: HardwareTrigger = copy.deepcopy(trigger_active)
            trigger_end: HardwareTrigger = copy.deepcopy(trigger_active)

            try:
                trigger_end.trigger_activation = {
                    TriggerActivation.RISING_EDGE: TriggerActivation.FALLING_EDGE,
                    TriggerActivation.FALLING_EDGE: TriggerActivation.RISING_EDGE,
                    TriggerActivation.ANY_EDGE: None,
                    TriggerActivation.LEVEL_LOW: None,
                    TriggerActivation.LEVEL_HIGH: None,
                }[trigger_end.trigger_activation]
            except KeyError:
                raise DriverFeatureNotImplemented(f'Unrecognized trigger activation: {trigger_end.trigger_activation}.')

        return trigger_start, trigger_end

    def set_acquisition_mode(
        self, sensor: Camera, acquisition_mode: Union[AcquisitionMode, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Set the acquisition mode for the specified sensor.

        Args:
            sensor (Camera): The camera object for which the acquisition mode is to be set.
            acquisition_mode (AcquisitionMode): The acquisition mode to be applied to the sensor.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound:
                If the sensor is currently not found.
            SensorInvalidState:
                If the sensor's state is not STARTED or CONNECTED.
            DriverStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            DriverFeatureNotImplemented:
                If the provided `acquisition_mode` is not implemented in the `_apply_acquisition_mode` method.
            AttributeError:
                If a required attribute cannot be set on the `node map`'s node map.
            AccessException:
                If there is an issue accessing or modifying specific attributes of the `node map`.

        Notes:
            - The method first tries to configure the frame rate settings if `max_rate` is specified.
            - If frame rate settings are not supported, it attempts to configure line rate settings.
            - The `_apply_acquisition_mode` method is used to apply additional acquisition settings.
            - Ensure that the `sensor` and `acquisition_mode` are correctly initialized and compatible with the method.
        """
        if not is_available(acquisition_mode):
            return

        sensor = self._sensor(sensor)

        was_started = sensor.state == SensorState.STARTED
        if was_started:
            self.stop(sensor)

        try:
            node_map = self._node_map(sensor)

            rate_limited = is_available(acquisition_mode.max_rate)
            try:
                node_map.AcquisitionFrameRateEnable.value = rate_limited
                if rate_limited:
                    # TODO: bounds check
                    node_map.AcquisitionFrameRate.value = acquisition_mode.max_rate
            except AttributeError:
                try:
                    # TODO: bounds check
                    node_map.AcquisitionLineRateEnable.value = rate_limited
                    if rate_limited:
                        node_map.AcquisitionLineRate.value = acquisition_mode.max_rate
                except AttributeError as e:
                    if is_available(acquisition_mode.max_rate):
                        raise SensorFeaturePermanentlyNotAvailable(e) from None
                except AccessException as e:
                    raise SensorFeatureTemporarilyNotAvailable(e) from None
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e) from None
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            # Apply the rest of the acquisition mode
            try:
                self._apply_acquisition_mode(acquisition_mode, node_map, sensor)
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e) from None
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e) from None
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            sensor.acquisition_mode = acquisition_mode

        finally:
            if was_started:
                self.start(sensor)

    @singledispatchmethod
    def _apply_acquisition_mode(self, acquisition_mode: AcquisitionMode, node_map, sensor: Camera) -> None:
        """Apply the specified acquisition mode to the image acquirer and camera.

        This method serves as an entry point and the default implementation for applying acquisition modes.
        Specific implementations are registered with this single dispatch method.

        Args:
            acquisition_mode (AcquisitionMode): The acquisition mode to apply.
            node_map: The image acquirer object that corresponds to the sensor
            that is changed.
            sensor (Camera): The camera object to which the acquisition settings will be applied.

        Raises:
            DriverStateException: if the driver is in an invalid state

            DriverFeatureNotImplemented: If this method is not overridden by a specific acquisition mode implementation.
            AttributeError: If a required attribute could not be set.
            AccessException: If there is an issue accessing or modifying specific attributes.
        """
        raise DriverFeatureNotImplemented

    @_apply_acquisition_mode.register
    def _apply_continuous_acquisition(
        self, acquisition_mode: ContinuousAcquisitionMode, node_map, sensor: Camera
    ) -> None:
        """Apply continuous acquisition mode to the image acquirer and camera.

        This method configures the image acquirer and camera for continuous acquisition mode. It sets
        the acquisition mode to 'Continuous' and disables all triggers associated with the camera.

        Args:
            acquisition_mode (AcquisitionMode): The acquisition mode to apply.
            node_map: The image acquirer object that corresponds to the sensor
            that is changed.
            sensor (Camera): The camera object to which the acquisition settings will be applied.

        Raises:
            DriverStateException: if the driver is in an invalid state

            AttributeError: If a required attribute could not be set.
            AccessException: If there is an issue accessing or modifying specific attributes.
        """
        node_map.AcquisitionMode.value = 'Continuous'
        sensor.acquisition_mode = acquisition_mode

        for trigger_name in filter(lambda a: a, sensor.camera_model.acquisition_trigger_names):
            node_map.TriggerSelector.value = trigger_name
            node_map.TriggerMode.value = 'Off'

    @_apply_acquisition_mode.register
    def _apply_trigger_acquisition(self, acquisition_mode: TriggerAcquisitionMode, node_map, sensor: Camera) -> None:
        """Apply trigger acquisition mode to the image acquirer and camera.

        This method configures the image acquirer and camera for trigger-based acquisition mode. It
        resets all existing triggers, applies new trigger settings, and configures burst properties if
        specified in the acquisition mode.

        Args:
            acquisition_mode (AcquisitionMode): The acquisition mode to apply.
            node_map: The image acquirer object that corresponds to the sensor
            that is changed.
            sensor (Camera): The camera object to which the acquisition settings will be applied.

        Raises:
            DriverStateException: if the driver is in an invalid state

            AttributeError: If a required attribute could not be set.
            AccessException: If there is an issue accessing or modifying specific attributes.
        """
        try:
            node_map.AcquisitionMode.value = 'SingleFrame'
        except Exception as e:
            if node_map.AcquisitionMode.value != 'SingleFrame':
                raise e from None

        # Reset all triggers
        for trigger_name in filter(lambda a: a, sensor.camera_model.acquisition_trigger_names):
            node_map.TriggerSelector.value = trigger_name
            node_map.TriggerMode.value = 'Off'

        # Set frame burst trigger
        self.apply_trigger(
            node_map, sensor.camera_model.frame_burst_start_trigger, acquisition_mode.trigger_frame_burst_start
        )
        self.apply_trigger(
            node_map, sensor.camera_model.frame_burst_stop_trigger, acquisition_mode.trigger_frame_burst_end
        )

        # Set frame burst properties
        if is_available(acquisition_mode.burst_frame_count):
            node_map.AcqusitionBurstFrameCount.value = acquisition_mode.burst_frame_count

        if is_available(acquisition_mode.burst_mode):
            node_map.BslAcquisitionBurstMode.value = acquisition_mode.burst_mode

        # Set frame trigger
        if sensor.camera_model.line_start_trigger is not None:
            self.apply_trigger(node_map, sensor.camera_model.line_start_trigger, acquisition_mode.trigger_start)
        else:
            self.apply_trigger(node_map, sensor.camera_model.frame_start_trigger, acquisition_mode.trigger_start)
        self.apply_trigger(node_map, sensor.camera_model.frame_stop_trigger, acquisition_mode.trigger_frame_end)

        sensor.acquisition_mode = acquisition_mode

    @staticmethod
    def apply_trigger(node_map, trigger_name: Optional[str], trigger: Optional[TriggerMode]) -> None:
        """Apply the specified trigger settings to the image acquirer of a specific sensor.

        If `trigger_name` is `None`, no changes are made. If `trigger` is `None`, the trigger is turned off.
        Otherwise, the trigger is enabled and configured with the specified source, delay, and activation settings.

        Args:
            node_map:
                The node map object of the sensor we would like to change.

            trigger_name (Optional[str]):
                The name of the trigger selector to apply. If `None`, no trigger settings are applied.

            trigger (Optional[TriggerMode]):
                The trigger mode to apply. If `None`, the trigger is turned off. If a valid
                `TriggerMode` instance is provided, the trigger is enabled and configured accordingly.

        Raises:
            DriverStateException: if the driver is in an invalid state

            ValueError:
                If the `trigger_activation` does not match any value in the predefined mapping.

            AttributeError:
                If a required attribute could not be set on the image acquirer’s node map.

            AccessException:
                If there is an issue accessing or modifying specific attributes of the image acquirer’s node map.
        """
        if trigger_name is None:
            return

        node_map.TriggerSelector.value = trigger_name
        if not is_available(trigger):
            node_map.TriggerMode.value = 'Off'
        else:
            node_map.TriggerMode.value = 'On'
            node_map.TriggerSource.value = trigger.trigger_source

            value: float
            try:
                value = node_map.TriggerDelay.value
            except (AttributeError, AccessException):
                value = 0.0

            trigger_delay: float = trigger.trigger_delay or 0.0
            try:
                node_map.TriggerDelay.value = trigger_delay
            except (AttributeError, AccessException) as e:
                if value != trigger_delay:
                    raise e from None

            if isinstance(trigger, HardwareTrigger):
                node_map.TriggerActivation.value = HarvestersDriver.serialize_trigger_activation(
                    trigger.trigger_activation
                )

    #
    # Functionality for requesting the current acquisition status
    #

    def compile_acquisition_status(self, sensor: Camera) -> str:
        """Compile an acquisition status string.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            Union[ExposureStatus, NotAvailableRead]: the acquisition status

        Raises:
            SensorNotFound: If the sensor or driver is in an invalid state
        """
        sensor = self._sensor(sensor)

        frame_burst_status: Union[FrameBurstStatus, NotAvailableRead] = self.fetch_frame_burst_status(sensor)
        frame_burst_string: str
        if is_available(frame_burst_status):
            try:
                frame_burst_string = {
                    FrameBurstStatus.WAIT_TRIGGER: 'waiting for trigger',
                    FrameBurstStatus.IDLE: 'idle',
                    FrameBurstStatus.ACTIVE: 'active',
                }[frame_burst_status]
            except KeyError:
                frame_burst_string = 'reported value not implemented'
        else:
            frame_burst_string = 'reporting not available'

        acquisition_status: Union[AcquisitionStatus, NotAvailableRead] = self.fetch_acquisition_status(sensor)
        acquisition_string: str
        if is_available(acquisition_status):
            try:
                acquisition_string = {
                    AcquisitionStatus.WAIT_TRIGGER: 'waiting for trigger',
                    AcquisitionStatus.IDLE: 'idle',
                    AcquisitionStatus.ACTIVE: 'active',
                }[acquisition_status]
            except KeyError:
                acquisition_string = 'reported value not implemented'
        else:
            acquisition_string = 'reporting not available'

        exposure_status: Union[ExposureStatus, NotAvailableRead] = self.fetch_exposure_status(sensor)
        exposure_string: str
        if is_available(exposure_status):
            try:
                exposure_string = {
                    ExposureStatus.WAIT_TRIGGER: 'waiting for trigger',
                    ExposureStatus.IDLE: 'idle',
                    ExposureStatus.ACTIVE: 'active',
                }[exposure_status]
            except KeyError:
                exposure_string = 'reported value not implemented'
        else:
            exposure_string = 'reporting not available'

        return f'Frame burst: {frame_burst_string}, Acquisition: {acquisition_string}, Exposure: {exposure_string}'

    def fetch_acquisition_status(self, sensor: Camera) -> Union[AcquisitionStatus, NotAvailableRead]:
        """Returns the acquisition status if possible.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            Union[ExposureStatus, NotAvailableRead]: the acquisition status

        Raises:
            SensorNotFound: If the sensor or driver is in an invalid state
        """
        sensor = self._sensor(sensor)

        active = self.retrieve_acquisition_status_acquisition_active(sensor)
        if is_available(active) and active:
            return AcquisitionStatus.ACTIVE

        idle = self.retrieve_acquisition_status_acquisition_idle(sensor)
        if is_available(idle) and idle:
            return AcquisitionStatus.IDLE

        frame_trigger_wait = self.retrieve_acquisition_status_frame_trigger_wait(sensor)
        if is_available(frame_trigger_wait) and frame_trigger_wait:
            return AcquisitionStatus.WAIT_TRIGGER

        line_trigger_wait = self.retrieve_acquisition_status_line_trigger_wait(sensor)
        if is_available(line_trigger_wait) and line_trigger_wait:
            return AcquisitionStatus.WAIT_TRIGGER

        return NotAvailableRead()

    def fetch_exposure_status(self, sensor: Camera) -> Union[ExposureStatus, NotAvailableRead]:
        """Returns the exposure status if possible.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            Union[ExposureStatus, NotAvailableRead]: the exposure status

        Raises:
            SensorNotFound: If the sensor or driver is in an invalid state
        """
        sensor = self._sensor(sensor)

        active = self.retrieve_acquisition_status_exposure_active(sensor)
        if is_available(active) and active:
            return ExposureStatus.ACTIVE

        idle = self.retrieve_acquisition_status_exposure_trigger_wait(sensor)
        if is_available(idle) and idle:
            return ExposureStatus.IDLE

        return NotAvailableRead()

    def fetch_frame_burst_status(self, sensor: Camera) -> Union[FrameBurstStatus, NotAvailableRead]:
        """Returns the frame burst status if possible.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            Union[ExposureStatus, NotAvailableRead]: the frame burst status

        Raises:
            SensorNotFound: If the sensor or driver is in an invalid state
        """
        sensor = self._sensor(sensor)

        active = self.retrieve_acquisition_status_frame_burst_active(sensor)
        if is_available(active) and active:
            return FrameBurstStatus.ACTIVE

        trigger_wait = self.retrieve_acquisition_status_frame_burst_trigger_wait(sensor)
        if is_available(trigger_wait) and trigger_wait:
            return FrameBurstStatus.WAIT_TRIGGER

        return NotAvailableRead()

    def retrieve_acquisition_status_acquisition_active(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether acquisition has started, i.e., whether the camera has received an
        Acquisition Start command.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the acquisition is currently active on the camera.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'AcquisitionActive')

    def retrieve_acquisition_status_acquisition_idle(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determine if the camera is currently idle.

        It is considered idle, if it is not currently exposing or reading out image data, and is therefore ready
        to receive an Acquisition Start command.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is currently idle.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'AcquisitionIdle')

    def retrieve_acquisition_status_exposure_active(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether the camera is currently exposing.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is currently exposing.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'ExposureActive')

    def retrieve_acquisition_status_exposure_trigger_wait(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether the camera is waiting for an Exposure Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is waiting for an Exposure Start trigger signal.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'ExposureTriggerWait')

    def retrieve_acquisition_status_frame_burst_active(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether the camera is currently acquiring a series of images (a "burst" of images).

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is currently acquiring a series of images (a "burst" of images).
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'FrameBurstActive')

    def retrieve_acquisition_status_frame_burst_trigger_wait(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether the camera is waiting for a Frame Burst Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is waiting for a Frame Burst Start trigger signal.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'FrameBurstTriggerWait')

    def retrieve_acquisition_status_frame_trigger_wait(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether the camera is waiting for a Frame Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is waiting for a Frame Start trigger signal.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'FrameTriggerWait')

    def retrieve_acquisition_status_line_trigger_wait(self, sensor: Camera) -> Union[bool, NotAvailableRead]:
        """Determines whether the camera is waiting for a Line Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            Union[bool, NotAvailable]: if the camera is waiting for a Line Start trigger signal.
        """
        sensor = self._sensor(sensor)
        return self._retrieve_acquisition_status(self._node_map(sensor), 'LineTriggerWait')

    @staticmethod
    def _retrieve_acquisition_status(node_map, status: str) -> Union[bool, NotAvailableRead]:
        """Fetch the current acquisition status of a specific feature.

        Arguments:
            node_map: The image acquirer associated to the camera

            status (str): the status selector string

        Returns:
            Union[bool, NotAvailableRead]: the acquisition status of the requested feature.
        """
        try:
            node_map.AcquisitionStatusSelector.value = status
            return node_map.AcquisitionStatus.value
        except (InvalidArgumentException, AttributeError):
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

    #
    # Triggering
    #

    # @abstractmethod
    def execute_software_trigger_frame(self, sensor: Camera, start: bool, timestamp: Optional[int] = None) -> None:
        """Execute a software start or stop trigger for triggering a frame.

        Arguments:
            sensor (Camera): The camera that is about to be triggered.

            start (bool): frame start / frame stop.

            timestamp (Optional[int]) timestamp of the software trigger.k:w


        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode, or because the camera
                                      does not support / is not configured with a stop trigger.
        """
        sensor = self._sensor(sensor)
        if not isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            raise InvalidArgumentException

        trigger = sensor.acquisition_mode.trigger_start if start else sensor.acquisition_mode.trigger_frame_end

        sensor.trigger_count = (sensor.trigger_count or 0) + 1

        if (
            not is_available(trigger)
            or trigger.trigger_source not in sensor.camera_model.software_trigger_sources_generic
        ):
            if not start:
                raise InvalidArgumentException('Sensor is not in triggered exposure mode.')

            node_map = self._node_map(sensor)
            for i in itertools.count(start=0):
                try:
                    node_map.TriggerSoftware.execute()
                    break
                except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                    ia = self._image_acquirer(sensor)
                    ia.start()
                    if i == 5:
                        # raise e from None
                        raise SensorLost(sensor)

        else:
            node_map = self._node_map(sensor)
            # node_map.TriggerSoftware.execute()
            self.execute_software_trigger(sensor, trigger)

        sensor.trigger_timestamp = timestamp

    def execute_software_trigger_frame_burst(self, sensor: Camera, start: bool) -> None:
        """Execute a software start or stop trigger for triggering a frame.

        Arguments:
            sensor (Camera): The camera that is about to be triggered.

            start (bool): frame start / frame stop.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode, or because the camera
                                      does not support / is not configured with a stop trigger.
        """
        sensor = self._sensor(sensor)
        mode: Optional[AcquisitionMode] = sensor.acquisition_mode
        if not isinstance(mode, TriggerAcquisitionMode):
            raise InvalidArgumentException('Sensor is not in TriggerAcquisitionMode.')

        self.execute_software_trigger(sensor, mode.trigger_frame_burst_start if start else mode.trigger_frame_burst_end)

    def execute_software_trigger_exposure(self, sensor: Camera, start: bool) -> None:
        """Execute te software start or stop trigger that controls exposure.

        Arguments:
            sensor (Camera): The camera that is about to be triggered.

            start (bool): frame exposure / stop exposure.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode, or because the camera
                                      does not support / is not configured with a stop trigger.
        """
        sensor = self._sensor(sensor)
        exposure: Optional[Exposure] = sensor.brightness.exposure
        if not isinstance(exposure, TriggeredExposure):
            raise InvalidArgumentException('Sensor is not in triggered exposure mode.')

        self.execute_software_trigger(
            sensor, exposure.trigger_exposure_start if start else exposure.trigger_exposure_end
        )

    def execute_software_trigger(
        self, sensor: Camera, trigger: Optional[Union[TriggerMode, str]], timestamp: Optional[int] = None
    ) -> None:
        """Execute te software start or stop trigger that controls exposure.

        Arguments:
            sensor (Camera): The camera that is about to be triggered.

            trigger (Optional[Union[TriggerMode, str]]): Trigger that is supposed to be executed.

            timestamp (Optional[int]): Trigger that is supposed to be executed.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is in an incompatible mode, or because the camera
                                      does not recognize / support the trigger that was provided.
        """
        sensor = self._sensor(sensor)
        if trigger is None:
            raise InvalidArgumentException('Cannot execute empty trigger.')

        trigger_source: str
        if isinstance(trigger, TriggerMode):
            trigger_source = trigger.trigger_source
        else:
            trigger_source = trigger

        if trigger_source not in sensor.camera_model.software_trigger_sources_generic:
            raise InvalidArgumentException(f'Trigger source {trigger_source} is not a software trigger source.')

        node_map = self._node_map(sensor)
        node_map.SequencerTriggerSelector.value = trigger_source
        node_map.SoftwareSignalPulse.execute()

    def configure_output_line_value(self, sensor: Camera, output_line_name: str, trigger_value: bool) -> None:
        """Set output line to high / low.

        Arguments:
            sensor (Camera): The camera that is about to be triggered.

            output_line_name (str): Name of the output line.

            trigger_value (bool): value the trigger should assume.

        Raises:
            InvalidState: if the driver is in an invalid state

            ResourceError: if the resource is unavailable

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            node_map.LineSelector.value = output_line_name
            # We could accidentally trigger multiple outputs, but don't want to open up this configuration rabbit hole.
            node_map.UserOutputSelector.value = node_map.LineSource.value
            node_map.UserOutputValue.Value = trigger_value
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    #
    # Regions Of Interest
    #

    def highlight_roi(self, sensor: Camera, roi: Optional[str] = None, highlight: bool = True) -> None:
        """Visually highlights a region of interest for configuration purpose.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

            roi (str): the name of the region of interest

            highlight (bool) whether to highlight the ROI in the image.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            SensorFeatureNotAvailable if the sensor does not support highlighting rois.

            SensorFeatureCurrentlyNotAvailable if the sensor currently does not support highlighting rois.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            if roi is not None:
                node_map.AutoFunctionROISelector.value = roi
            node_map.AutoFunctionROIHighlight.value = highlight
        except (AttributeError, LogicalErrorException):
            raise SensorFeaturePermanentlyNotAvailable from None
        except AccessException:
            raise SensorFeatureTemporarilyNotAvailable
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def fetch_roi(self, sensor: Camera, roi: str) -> Union[Roi, NotAvailable]:
        """Retrieve roi from camera.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

            roi (str): the name of the region of interest

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        try:
            node_map.AutoFunctionROISelector.value = roi
            w = node_map.Width.value
            h = node_map.Height.value
            x = node_map.OffsetX.value
            y = node_map.OffsetY.value
            return Roi(x=x, y=y, width=w, height=h)

        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def configure_roi(self, sensor: Camera, name: str, roi: Optional[Union[Roi, NotAvailable]]) -> None:
        """Configure an auto function ROI.

        The settings for the auto function rois are independent of the settings for the image roi.
        There should be no effect of the general ROI onto the configured auto rois.

        Attributes:
            sensor (Camera): The sensor, to which the roi settings are to be applied.

            name (str): The name of the region of interest that is to be configured.

            roi: Union[Roi, NotAvailable]: The region of interest that is to be configured.
                                If this value is None, defaults to the entire image.

                                Attention: Don't use None if you configured a camera crop,
                                and want to use the same pixels as in the camera crop.
                                Use `sensor.crop` as argument instead!

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is not implemented

            InvalidArgumentException if the ranges are not valid.
        """
        sensor = self._sensor(sensor)
        if isinstance(roi, NotAvailableRead):
            return

        if isinstance(roi, NotAvailableWrite):
            raise SensorFeaturePermanentlyNotAvailable

        if not is_available(roi):
            if roi is not None:
                raise DriverFeatureImplementationError

        node_map = self._node_map(sensor)

        try:
            node_map.AutoFunctionROISelector.value = name
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        # Auto-configure ROI to the entire image.
        if roi is None:
            try:
                roi = Roi(
                    x=node_map.AutoFunctionROIOffsetX.min,
                    y=node_map.AutoFunctionROIOffsetY.min,
                    width=node_map.AutoFunctionROIWidth.max,
                    height=node_map.AutoFunctionROIHeight.max,
                )
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        try:
            previous_x = node_map.AutoFunctionROIOffsetX.value
            previous_y = node_map.AutoFunctionROIOffsetY.value
            previous_w = node_map.AutoFunctionROIWidth.value
            previous_h = node_map.AutoFunctionROIHeight.value
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        # @dev: If not resetting offsetX or Width to their minimal values, an exception can
        #       occur even if the pair (x, width) is valid. See set_crop for details.
        try:
            node_map.AutoFunctionROIOffsetX.value = 0
            node_map.AutoFunctionROIWidth.value = roi.width
            node_map.AutoFunctionROIOffsetX.value = roi.x

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)

        except OutOfRangeException as e:
            # Thrown if the width is not dividable by the increment, or w out of range.
            try:
                node_map.AutoFunctionROIOffsetX.value = previous_x
                node_map.AutoFunctionROIWidth.value = previous_w
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except OutOfRangeException:
                raise InvalidArgumentException(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        try:
            node_map.AutoFunctionROIOffsetY.value = 0
            node_map.AutoFunctionROIHeight.value = roi.height
            node_map.AutoFunctionROIOffsetY.value = roi.y

        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        except OutOfRangeException as e:
            # Thrown if the width is not dividable by the increment, or h out of range.
            try:
                node_map.AutoFunctionROIOffsetY.value = previous_y
                node_map.AutoFunctionROIHeight.value = previous_h
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except OutOfRangeException:
                raise InvalidArgumentException(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        sensor.rois[name] = roi

    def __configure_roi_for_feature(
        self,
        sensor: Camera,
        node_map,
        new_roi_configuration: Optional[Union[str, Roi]],
        attribute: str,
        old_roi_configuration: Union[str, Roi],
    ) -> None:
        """Configure exactly one region of interest as the active region of interest for a particular feature.

        Attributes:
            sensor (Camera): current camera.

            node_map: node map associated to the current camera.

            new_roi_configuration (Union[str, Roi]): roi configuration that is to be applied.

            old_roi_configuration (Union[str, Roi]): roi configuration that used to be applied.

            attribute (str): Attribute in the image acquirer's node map that determines the feature.

        Raises:
            DriverStateException: if the driver is in an invalid state

            NoFreeRois: if a roi is required, but no region of interest is free.

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is not implemented

            InvalidArgumentException if the ranges are not valid.
        """
        sensor = self._sensor(sensor)
        selected_roi: Optional[str] = None
        if isinstance(new_roi_configuration, str):
            selected_roi = new_roi_configuration
        elif isinstance(new_roi_configuration, Roi):
            if isinstance(old_roi_configuration, str):
                selected_roi = old_roi_configuration

        if selected_roi is None:
            try:
                selected_roi = next(iter(sensor.get_free_roi()))
            except StopIteration:
                raise NoFreeRois.create()

        for roi in sensor.get_free_roi():
            is_selected: bool = roi == selected_roi
            node_map.AutoFunctionROISelector.value = roi
            node_map.__getattr__(attribute).value = is_selected

        if isinstance(new_roi_configuration, Roi) or new_roi_configuration is None:
            self.configure_roi(sensor, selected_roi, new_roi_configuration)

    #
    # Brightness
    #

    def fetch_brightness(self, sensor: Camera) -> Union[Brightness, NotAvailable]:
        """Return a camera's brightness settings.

        Returns:
            Optional[Brightness] the camera's brightness settings if they could be retrieved.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            DriverFeatureNotImplemented if the feature is not available

            ValueError if a retrieved value does not match the expectations
        """
        sensor = self._sensor(sensor)
        gain: Optional[Gain] = self.__fetch_gain(sensor)
        exposure: Optional[Exposure] = self.__fetch_exposure(sensor)

        node_map = self._node_map(sensor)
        auto_roi_configuration: Optional[Union[str, NotAvailable]] = None
        for roi in sensor.camera_model.rois:
            try:
                node_map.AutoFunctionROISelector.value = roi
                if node_map.AutoFunctionROIUseBrightness.value:
                    auto_roi_configuration = roi
            except AttributeError:
                if not is_available(auto_roi_configuration):
                    auto_roi_configuration = NotSupported()
            except AccessException:
                if not is_available(auto_roi_configuration):
                    auto_roi_configuration = TemporarilyNotAvailable()
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        auto_target_brightness: Union[float, NotAvailable]
        try:
            auto_target_brightness = node_map.AutoTargetBrightness.value
        except AttributeError:
            auto_target_brightness = NotSupported()
        except AccessException:
            auto_target_brightness = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        auto_target_brightness_damping: Union[float, NotAvailable]
        try:
            auto_target_brightness_damping = node_map.AutoTargetBrightnessDamping.value
        except AttributeError:
            auto_target_brightness_damping = NotSupported()
        except AccessException:
            auto_target_brightness_damping = TemporarilyNotAvailable()

        auto_function_profile: Union[str, NotAvailable]
        try:
            auto_function_profile = node_map.AutoFunctionProfile.value
        except AttributeError:
            auto_function_profile = NotSupported()
        except AccessException:
            auto_function_profile = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        auto_backlight_compensation: Union[float, NotAvailable]
        try:
            auto_backlight_compensation = node_map.AutoBacklightCompensation.value
        except AttributeError:
            auto_backlight_compensation = NotSupported()
        except AccessException:
            auto_backlight_compensation = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        auto_gray_value_adjustment: Union[float, NotAvailable]
        try:
            auto_gray_value_adjustment = node_map.GrayValueAdjustmentDampingAbs.value
        except AttributeError:
            auto_gray_value_adjustment = NotSupported()
        except AccessException:
            auto_gray_value_adjustment = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if not is_available(auto_target_brightness_damping):
            auto_target_brightness_damping = auto_gray_value_adjustment

        black_level: Union[BlackLevel, NotAvailable]
        try:
            black_level = ManualBlackLevel(node_map.BlackLevel.value)
        except AttributeError:
            black_level = NotSupported()
        except AccessException:
            black_level = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        return Brightness(
            exposure=exposure,
            gain=gain,
            black_level=black_level,
            auto_roi_configuration=auto_roi_configuration,
            auto_target_brightness=auto_target_brightness,
            auto_target_brightness_damping=auto_target_brightness_damping,
            auto_function_profile=auto_function_profile,
            auto_backlight_compensation=auto_backlight_compensation,
        )

    def set_brightness(self, sensor: Camera, brightness: Union[Brightness, NotAvailable], bounds_check: bool = False):
        """Configure brightness.

        Arguments:
            sensor (Camera): The camera to which we apply the settings.

            brightness (Brightness) The brightness configuration.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        if not is_available(brightness):
            return

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        try:
            self.__configure_roi_for_feature(
                sensor,
                node_map,
                brightness.auto_roi_configuration,
                'AutoFunctionROIUseBrightness',
                sensor.brightness.auto_roi_configuration if sensor.brightness.auto_function_enabled else None,
            )
        except NoFreeRois as e:
            if is_available(brightness.auto_roi_configuration):
                raise e from None
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e) from None

        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e) from None

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if is_available(brightness.auto_target_brightness):
            try:
                target: float
                if bounds_check:
                    target = min(
                        max(brightness.auto_target_brightness, node_map.AutoTargetBrightness.min),
                        node_map.AutoTargetBrightness.max,
                    )
                else:
                    target = brightness.auto_target_brightness

                node_map.AutoTargetBrightness.value = target

            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e) from None

            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e) from None

            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        if is_available(brightness.auto_target_brightness_damping):
            try:
                target: float
                if bounds_check:
                    target = min(
                        max(brightness.auto_target_brightness_damping, node_map.AutoTargetBrightnessDamping.min),
                        node_map.AutoTargetBrightnessDamping.max,
                    )
                else:
                    target = brightness.auto_target_brightness

                node_map.AutoTargetBrightnessDamping.value = target

            except AttributeError:
                if is_available(brightness.auto_target_brightness_damping):
                    try:
                        target: float
                        if bounds_check:
                            target = min(
                                max(
                                    brightness.auto_target_brightness_damping,
                                    node_map.GrayValueAdjustmentDampingAbs.min,
                                ),
                                node_map.GrayValueAdjustmentDampingAbs.max,
                            )
                        else:
                            target = brightness.auto_target_brightness

                        node_map.GrayValueAdjustmentDampingAbs.value = target
                    except AttributeError as e:
                        raise SensorFeaturePermanentlyNotAvailable(e) from None

                    except AccessException as e:
                        raise SensorFeatureTemporarilyNotAvailable(e) from None

            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e) from None

            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        if is_available(brightness.auto_function_profile):
            try:
                node_map.AutoFunctionProfile.value = brightness.auto_function_profile

            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e) from None

            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e) from None

            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        if is_available(brightness.auto_backlight_compensation):
            try:
                target: float
                if bounds_check:
                    target = min(
                        max(brightness.auto_target_brightness_damping, node_map.AutoBacklightCompensation.min),
                        node_map.AutoBacklightCompensation.max,
                    )
                else:
                    target = brightness.auto_target_brightness

                node_map.AutoBacklightCompensation.value = target

            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e) from None

            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e) from None

            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        if is_available(brightness.black_level):
            if isinstance(brightness.black_level, ManualBlackLevel):
                try:
                    node_map.BlackLevel.value = brightness.black_level.value

                except AttributeError as e:
                    raise SensorFeaturePermanentlyNotAvailable(e) from None

                except AccessException as e:
                    raise SensorFeatureTemporarilyNotAvailable(e) from None

                except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                    raise SensorLost(sensor)

        try:
            if is_available(brightness.gain):
                self.__set_gain(sensor, brightness.gain, bounds_check=bounds_check)
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e) from None
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e) from None
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            if is_available(brightness.exposure):
                self.__set_exposure(sensor, brightness.exposure, bounds_check=bounds_check)
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e) from None
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e) from None

        sensor.brightness = brightness

    #
    # Gain
    #

    def __fetch_gain(self, sensor: Camera) -> Union[Gain, NotAvailable]:
        """Fetch the camera's current gain configuration.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            bool: the camera's gain configuration if it could be retrieved.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            DriverFeatureNotImplemented if not implemented
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        gain_selector_digital: Optional[bool]
        try:
            gain_selector: Optional[str] = node_map.GainSelector.value
            if gain_selector not in ('DigitalAll', 'AnalogAll'):
                raise ValueError(f'Unknown gain selector: {gain_selector}')
            gain_selector_digital = gain_selector == 'DigitalAll'
        except (AttributeError, AccessException, ValueError):
            gain_selector_digital = None

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        auto_gain: bool
        temporary: bool
        try:
            value = node_map.GainAuto.value
            auto_gain = value in ('Continuous', 'Once')
            temporary = value == 'Once'

        except (AttributeError, AccessException):
            auto_gain = False
            temporary = False

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if not auto_gain:
            try:
                return ManualGain(gain_selector_digital, node_map.Gain.value)
            except AttributeError:
                return NotSupported()
            except AccessException:
                return TemporarilyNotAvailable()
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        upper_limit: Union[float, int, NotAvailable]
        try:
            upper_limit = node_map.AutoGainUpperLimit.value
        except AttributeError:
            upper_limit = NotSupported()
        except AccessException:
            upper_limit = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        lower_limit: Union[float, int, NotAvailable]
        try:
            lower_limit = node_map.AutoGainLowerLimit.value
        except AttributeError:
            lower_limit = NotSupported()
        except AccessException:
            lower_limit = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        return AutoGain(gain_selector_digital, upper_limit=upper_limit, lower_limit=lower_limit, temporary=temporary)

    def __set_gain(self, sensor: Camera, gain: Gain, bounds_check: bool = False) -> None:
        """Set the camera gain.

        The Gain camera feature allows you to increase the brightness of the images output by the camera.
        Increasing the gain increases all pixel values of the image.

        There are two ways of specifying the gain
            - automatic configuration
            - manual configuration by setting the gain value in [dB] manually.
              The minimum and maximum parameter values vary depending on the camera model,
              the pixel format chosen, and the Binning settings.

        For applying the gain, there are sometimes two options:
            - AnalogAll: Selects the analog gain control
            - DigitalAll: Selects the digital gain control


        Analog gain is applied before the signal from the camera sensor is converted into digital values.
        Digital gain is applied after the conversion, i.e., it is basically a multiplication of the digitized values.

        Arguments:
            sensor (Camera): The camera to which we apply the gain settings.

            gain (Gain): The gain that is to be applied.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            DriverFeatureNotImplemented if not implemented

            NoFreeRois: if a roi is required, but no region of interest is free.

            FeatureUnavailable if the feature is not implemented

            InvalidArgumentException if the ranges are not valid.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        if is_available(gain.digital):
            node_map.GainSelector.value = {True: 'DigitalAll', False: 'AnalogAll'}[gain.digital]

        if isinstance(gain, ManualGain):
            try:
                node_map.GainAuto.value = 'Off'
            except AttributeError:
                ...
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            gain_value = gain.value
            if bounds_check:
                gain_value = min(max(gain_value, node_map.Gain.min), node_map.Gain.max)
            node_map.Gain.value = gain_value

        elif isinstance(gain, AutoGain):
            try:
                node_map.GainAuto.value = 'Once' if gain.temporary else 'Continuous'
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            upper: Union[int, float]
            if not is_available(gain.upper_limit):
                upper = node_map.AutoGainUpperLimit.max
            else:
                upper = gain.upper_limit
                if bounds_check:
                    upper = min(max(upper, node_map.AutoGainUpperLimit.min), node_map.AutoGainUpperLimit.max)
            node_map.AutoGainUpperLimit.value = upper

            lower: Union[int, float]
            if not is_available(gain.lower_limit):
                lower = node_map.AutoGainLowerLimit.min
            else:
                lower = gain.upper_limit
                if bounds_check:
                    lower = min(max(lower, node_map.AutoGainLowerLimit.min), node_map.AutoGainLowerLimit.max)
            node_map.AutoGainLowerLimit.value = lower

        else:
            raise DriverFeatureNotImplemented

        sensor.brightness.gain = gain

    #
    # Exposure
    #

    def __fetch_exposure(self, sensor: Camera) -> Union[Exposure, NotAvailable]:
        """Fetch the camera's current exposure configuration.

        Arguments:
            sensor (Camera): The camera from which we gather the information.

        Returns:
            bool: the camera's exposure configuration if it could be retrieved.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            DriverFeatureNotImplemented if not implemented

            ValueError if a retrieved value does not match the expectations
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        camera_model: CameraModel = sensor.camera_model

        # Fetch the exposure mode
        exposure_mode: Optional[str]
        try:
            exposure_mode = node_map.ExposureMode.value
        except (AttributeError, AccessException):
            exposure_mode = None
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        # Fetch the trigger modes
        trigger_modes: dict[Optional[str], Optional[TriggerMode]] = {None: None}
        for trigger_name in filter(lambda a: a, sensor.camera_model.exposure_trigger_names):
            trigger_information = HarvestersDriver.fetch_trigger(node_map, camera_model, trigger_name)
            if is_available(trigger_information) and trigger_information[1]:
                trigger_modes[trigger_name] = trigger_information[0]
            else:
                trigger_modes[trigger_name] = None

        trigger_exposure_start, trigger_exposure_end = HarvestersDriver.determine_trigger_modes(
            trigger_modes[camera_model.exposure_start_trigger],
            trigger_modes[camera_model.exposure_stop_trigger],
            trigger_modes[camera_model.exposure_active_trigger],
        )

        auto: bool
        temporary: bool
        try:
            value = node_map.ExposureAuto.value
            auto = value in ('Continuous', 'Once')
            temporary = value == 'Once'

        except (AttributeError, AccessException):
            auto = False
            temporary = False
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if auto:
            if is_available(trigger_exposure_start) or is_available(trigger_exposure_end):
                raise ValueError

            upper_limit: Optional[Union[float, int]]
            try:
                upper_limit = node_map.AutoExposureTimeUpperLimit.value
            except AttributeError:
                upper_limit = None
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            lower_limit: Optional[Union[float, int]]
            try:
                lower_limit = node_map.AutoExposureTimeLowerLimit.value
            except AttributeError:
                lower_limit = None
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            return AutoExposure(upper_limit=upper_limit, lower_limit=lower_limit, temporary=temporary)

        # Timed exposure. For keyence exposure mode is not implemented, None means timed exposure
        if exposure_mode == 'Timed' or exposure_mode is None:
            if auto:
                raise ValueError

            exposure_time: Optional[Union[float, int]]
            try:
                exposure_time = node_map.ExposureTime.value
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            return TimedExposure(exposure_time)

        # Hardware triggered exposure.
        if exposure_mode == 'TriggerWidth':
            if not is_available(trigger_exposure_start):
                raise ValueError

            if not isinstance(trigger_exposure_start, HardwareTrigger):
                raise ValueError

            if trigger_exposure_start.trigger_activation not in (
                TriggerActivation.RISING_EDGE,
                TriggerActivation.FALLING_EDGE,
            ):
                raise ValueError

            if is_available(trigger_exposure_end):
                raise ValueError

        elif exposure_mode == 'TriggerControlled':
            if not is_available(trigger_exposure_start):
                raise ValueError

            if not isinstance(trigger_exposure_start, HardwareTrigger):
                raise ValueError

            if trigger_exposure_start.trigger_activation not in (
                TriggerActivation.RISING_EDGE,
                TriggerActivation.FALLING_EDGE,
            ):
                raise ValueError

            if not is_available(trigger_exposure_end):
                raise ValueError

            if not isinstance(trigger_exposure_end, HardwareTrigger):
                raise ValueError

            if trigger_exposure_end.trigger_activation not in (
                TriggerActivation.RISING_EDGE,
                TriggerActivation.FALLING_EDGE,
            ):
                raise ValueError

        else:
            raise DriverFeatureNotImplemented(f'Exposure mode "{exposure_mode}" not implemented.')

        start_delay: Optional[float]
        try:
            start_delay = node_map.BslExposureStartDelay.value
        except AttributeError:
            start_delay = None
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        overlap_time_max: Optional[float]
        try:
            overlap_time_max = node_map.BslExposureStartDelay.value
        except AttributeError:
            overlap_time_max = None
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        return TriggeredExposure(
            trigger_exposure_start=trigger_exposure_start,
            trigger_exposure_end=trigger_exposure_end,
            start_delay=start_delay,
            overlap_time_max=overlap_time_max,
        )

    def __set_exposure(self, sensor: Camera, exposure: Exposure, bounds_check: bool = False) -> None:
        """Set the camera exposure.

        Arguments:
            sensor (Camera): The camera to which we apply the exposure settings.

            exposure (Exposure): The exposure that is to be applied.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            DriverFeatureNotImplemented if not implemented

            NoFreeRois: if a roi is required, but no region of interest is free.

            FeatureUnavailable if the feature is not implemented

            InvalidArgumentException if the ranges are not valid.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        if isinstance(exposure, TimedExposure):
            try:
                try:
                    node_map.ExposureMode.value = 'Timed'
                except AttributeError:
                    pass

                try:
                    node_map.ExposureAuto.value = 'Off'
                except AttributeError:
                    pass

                exposure_time = exposure.time
                if is_available(exposure_time):
                    if bounds_check:
                        exposure_time = min(max(exposure_time, node_map.ExposureTime.min), node_map.ExposureTime.max)

                    if isinstance(node_map.ExposureTime.value, int):
                        exposure_time = int(exposure_time)

                    node_map.ExposureTime.value = exposure_time
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        elif isinstance(exposure, AutoExposure):
            try:
                node_map.ExposureMode.value = 'Timed'
                node_map.ExposureAuto.value = 'Once' if exposure.temporary else 'Continuous'

                upper: Union[int, float]
                if not is_available(exposure.upper_limit):
                    upper = node_map.AutoExposureTimeUpperLimit.max
                else:
                    upper = exposure.upper_limit
                    if bounds_check:
                        upper = min(
                            max(upper, node_map.AutoExposureTimeUpperLimit.min), node_map.AutoExposureTimeUpperLimit.max
                        )
                node_map.AutoExposureTimeUpperLimit.value = upper

                lower: Union[int, float]
                if not is_available(exposure.lower_limit):
                    lower = node_map.AutoExposureTimeLowerLimit.min
                else:
                    lower = exposure.upper_limit
                    if bounds_check:
                        lower = min(
                            max(lower, node_map.AutoExposureTimeLowerLimit.min), node_map.AutoExposureTimeLowerLimit.max
                        )
                node_map.AutoExposureTimeLowerLimit.value = lower
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        elif isinstance(exposure, TriggeredExposure):
            try:
                self.apply_trigger(
                    node_map, sensor.camera_model.exposure_start_trigger, exposure.trigger_exposure_start
                )

                if not is_available(exposure.trigger_exposure_end):
                    node_map.ExposureMode.value = 'TriggerWidth'

                else:
                    node_map.ExposureMode.value = 'TriggerControlled'

                self.apply_trigger(node_map, sensor.camera_model.exposure_stop_trigger, exposure.trigger_exposure_end)

                if is_available(exposure.start_delay):
                    node_map.BslExposureStartDelay.value = exposure.start_delay

                if is_available(exposure.overlap_time_max):
                    node_map.ExposureOverlapTimeMax.value = exposure.overlap_time_max

            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        else:
            raise DriverFeatureNotImplemented(f'Exposure mode {type(exposure).__name__} not implemented')

    #
    # White balance
    #

    def fetch_white_balance(self, sensor: Camera) -> Union[WhiteBalance, NotAvailable]:
        """Return the white balance settings.

        Arguments:
            sensor (Camera): The sensor from which we gather the white balance.

        Returns:
            Optional[float]: the camera's white balance.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        camera_model: CameraModel = sensor.camera_model

        auto: bool
        temporary: bool
        try:
            balance_white = node_map.BalanceWhiteAuto.value
            temporary = balance_white == 'Once'
            auto = balance_white == 'Continuous'
            auto = auto or temporary
        except (AttributeError, AccessException):
            auto = False
            temporary = False
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        auto_roi_configuration: Optional[Union[str, NotAvailable]] = None
        for roi_name in camera_model.rois:
            try:
                node_map.AutoFunctionROISelector.value = roi_name
                if node_map.AutoFunctionROIUseWhiteBalance.value:
                    auto_roi_configuration = roi_name
            except AttributeError:
                if not is_available(auto_roi_configuration):
                    auto_roi_configuration = NotSupported()
            except AccessException:
                if not is_available(auto_roi_configuration):
                    auto_roi_configuration = TemporarilyNotAvailable()
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        if auto:
            # If there is no configured auto ROI, return the error that we received while we attemptedto determine the
            # active auto roi.
            if not is_available(auto_roi_configuration):
                return TemporarilyNotAvailable() if not is_available(auto_roi_configuration) else auto_roi_configuration

            adjustment_damping: Union[float, NotAvailable]
            try:
                adjustment_damping = node_map.BalanceWhiteAdjustmentDampingAbs.value

            except AttributeError:
                adjustment_damping = NotSupported()
            except AccessException:
                adjustment_damping = TemporarilyNotAvailable()
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            return AutoWhiteBalance(auto_roi_configuration, adjustment_damping=adjustment_damping, temporary=temporary)

        try:
            white_balance_by_channel: dict[str, float] = {}
            for channel in ('Red', 'Green', 'Blue'):
                node_map.BalanceRatioSelector.value = channel
                white_balance_by_channel[channel] = node_map.BalanceRatio.value

            return ManualWhiteBalance(white_balance_by_channel)

        except (AttributeError, ErrorException):
            # The keyence cameras thor error exceptions when attributes are not found..
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_white_balance(
        self, sensor: Camera, white_balance: Union[WhiteBalance, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the white balance settings.

        Arguments:
            sensor (Camera): The sensor to which apply the white balance settings

            white_balance (WhiteBalance): The white balance settings

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        if not is_available(white_balance):
            return

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        if isinstance(white_balance, AutoWhiteBalance):
            try:
                self.__configure_roi_for_feature(
                    sensor,
                    node_map,
                    white_balance.roi_configuration,
                    'AutoFunctionROIUseWhiteBalance',
                    sensor.white_balance.roi_configuration
                    if isinstance(sensor.white_balance, AutoWhiteBalance)
                    else None,
                )

                node_map.BalanceWhiteAuto.value = 'Once' if white_balance.temporary else 'Continuous'

                adjustment_damping: Union[int, float]
                if not is_available(white_balance.adjustment_damping):
                    adjustment_damping = node_map.BalanceWhiteAdjustmentDampingAbs.min
                else:
                    adjustment_damping = white_balance.adjustment_damping
                    if bounds_check:
                        adjustment_damping = min(
                            max(adjustment_damping, node_map.BalanceWhiteAdjustmentDampingAbs.min),
                            node_map.BalanceWhiteAdjustmentDampingAbs.max,
                        )
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)

            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)

            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            try:
                node_map.BalanceWhiteAdjustmentDampingAbs.value = adjustment_damping
            except AttributeError as e:
                if is_available(white_balance.adjustment_damping):
                    raise SensorFeaturePermanentlyNotAvailable(e)

            except AccessException as e:
                if is_available(white_balance.adjustment_damping):
                    raise SensorFeatureTemporarilyNotAvailable(e)

            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        else:
            try:
                node_map.BalanceWhiteAuto.value = 'Off'
            except AttributeError:
                ...
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

            if isinstance(white_balance, ManualWhiteBalance):
                white_balance_channels = copy.copy(white_balance.channels)
                for channel in ('Red', 'Green', 'Blue'):
                    white_balance_channels[channel] = 1.0

                for channel, value in white_balance_channels.items():
                    try:
                        node_map.BalaneRatioSelector.value = channel

                        if bounds_check:
                            value = min(max(value, node_map.BalaneRatio.min), node_map.BalaneRatio.max)
                        node_map.BalaneRatio.value = value
                    except AttributeError as e:
                        raise SensorFeaturePermanentlyNotAvailable(
                            f'Value "{value}" could not be applied for channel "{channel}"; {e}.'
                        )
                    except AccessException as e:
                        raise SensorFeatureTemporarilyNotAvailable(e)
                    except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                        raise SensorLost(sensor)

            elif is_available(white_balance):
                raise DriverFeatureNotImplemented(
                    f'White balance mode f"{type(white_balance).__name__}" not implemented.'
                )

        sensor.white_balance = white_balance

    #
    # Binning
    #

    def fetch_binning(self, sensor: Camera) -> Union[Binning, NotAvailable]:
        """Return the binning configuration.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[Binning]: the camera's binning configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            DriverFeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        # Fetch optional values
        try:
            horizontal_mode = node_map.BinningHorizontalMode.value
        except AttributeError:
            horizontal_mode = NotSupported()
        except AccessException:
            horizontal_mode = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            vertical_mode = node_map.BinningVerticalMode.value
        except AttributeError:
            vertical_mode = NotSupported()
        except AccessException:
            vertical_mode = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            return Binning(
                horizontal=node_map.BinningHorizontal.value,
                horizontal_mode=horizontal_mode,
                vertical=node_map.BinningVertical.value,
                vertical_mode=vertical_mode,
            )
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_binning(
        self, sensor: Camera, binning: Optional[Union[Binning, NotAvailable]] = None, bounds_check: bool = False
    ) -> None:
        """Apply the binning configuration.

        Arguments:
            sensor (Camera): The sensor to which apply the binning configuration.

            binning (Optional[Binning]): The binning configuration or None to disable binning.


            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        if not is_available(binning) and binning is not None:
            return

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        if bounds_check:
            raise DriverFeatureNotImplemented('bounds check is currently not implemented for binning.')

        binning = Binning(1, NotAvailable(), 1, NotAvailable()) if not is_available(binning) else binning

        # Apply values. If binning is deactivated, throw no exception if the feature is not available.

        try:
            node_map.BinningHorizontal.value = binning.horizontal
        except AttributeError as e:
            if node_map.BinningHorizontal.value != 1:
                raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if is_available(binning.horizontal_mode):
            try:
                node_map.BinningHorizontalMode.value = binning.horizontal_mode
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        try:
            node_map.BinninVertical.value = binning.vertical
        except AttributeError as e:
            if node_map.BinninVertical.value != 1:
                raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if is_available(binning.vertical_mode):
            try:
                node_map.BinningVerticalMode.value = binning.vertical_mode
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        sensor.binning = binning

    #
    # Blooming reduction
    #

    def fetch_blooming_reduction(self, sensor: Camera) -> Union[bool, NotAvailable]:
        """Return the blooming reduction configuration.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[bool]: the camera's blooming reduction configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            return node_map.BslBloomingReductionEnable.value
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()

    def refresh_blooming_reduction(self, sensor: Camera) -> None:
        """Retrieve and apply the blooming reduction configuration.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        sensor.blooming_reduction = self.fetch_blooming_reduction(sensor)

    def set_blooming_reduction(
        self, sensor: Camera, blooming_reduction: Union[bool, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the blooming reduction configuration.

        Arguments:
            sensor (Camera): The sensor to which apply the blooming reduction configuration.

            blooming_reduction (bool): The blooming reduction configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        if not is_available(blooming_reduction):
            return

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            node_map.BslBloomingReductionEnable.value = blooming_reduction
        except AttributeError as e:
            if sensor.blooming_reduction:
                raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        sensor.blooming_reduction = blooming_reduction

    #
    # Brightness and contrast
    #

    def fetch_contrast_adjustments(self, sensor: Camera) -> Union[ContrastAdjustments, NotAvailable]:
        """Return the contrast adjustment configuration.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[bool]: the T's contrast adjustment configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        contrast_mode: Union[str, NotAvailable]
        try:
            contrast_mode = node_map.BslContrastMode.value
        except AttributeError:
            contrast_mode = NotSupported()
        except AccessException:
            contrast_mode = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            return ContrastAdjustments(
                contrast_mode=contrast_mode,
                brightness_pivot=node_map.BslBrightness.value,
                contrast_pivot=node_map.BslContrast.value,
            )
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_contrast_adjustments(
        self, sensor: Camera, contrast_adjustments: Union[ContrastAdjustments, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the contrast adjustment configuration.

        Arguments:
            sensor (Camera): The sensor to which apply the configuration.

            contrast_adjustments (bool): The contrast adjustment configuration.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        if not is_available(contrast_adjustments):
            return
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        if not is_available(contrast_adjustments):
            brightness, contrast, mode = 0.0, 0.0, None
        else:
            brightness, contrast, mode = (
                contrast_adjustments.brightness_pivot,
                contrast_adjustments.contrast_pivot,
                contrast_adjustments.contrast_mode,
            )

        if is_available(mode):
            try:
                node_map.BslContrastMode.value = mode
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        try:
            node_map.BslBrightness.value = brightness
        except AttributeError as e:
            if brightness:
                raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            node_map.BslContrast.value = contrast
        except AttributeError as e:
            if contrast:
                raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        sensor.contrast_adjustments = contrast_adjustments

    #
    # OperationMode
    #

    def fetch_operation_mode(self, sensor: Camera) -> Union[str, NotAvailable]:
        """Return the operation mode.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[bool]: The sensor's operation mode.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            return node_map.CameraOperationMode.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_operation_mode(
        self, sensor: Camera, operation_mode: Union[str, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the operation mode.

        Arguments:
            sensor (Camera): The sensor to which apply the configuration.

            operation_mode (str): the new operation mode

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        if not is_available(operation_mode):
            return

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            node_map.CameraOperationMode.value = operation_mode
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)
        sensor.operation_mode = operation_mode

    #
    # Color adjustments
    #

    def fetch_color_adjustment(self, sensor: Camera) -> Union[ColorAdjustment, NotAvailable]:
        """Return the color adjustment.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[bool]: The sensor's color adjustment.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        camera = node_map

        color_adjustments: dict[str, Union[ColorAdjustmentConfiguration, NotAvailable]] = {}
        for color in ('red', 'green', 'blue', 'yellow', 'cyan', 'magenta'):
            try:
                camera.ColorAdjustmentSelector.value = color
                hue, saturation = camera.ColorAdjustmentHue.value, camera.ColorAdjustmentSaturation.value
                color_adjustments[color] = ColorAdjustmentConfiguration(
                    hue_rotation=hue, saturation_adjustment=saturation
                )
            except AttributeError:
                color_adjustments[color] = NotSupported()
            except AccessException:
                color_adjustments[color] = TemporarilyNotAvailable()
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        hue: Union[float, int, NotAvailable]
        try:
            hue = camera.Hue.value
        except (AttributeError, AccessException):
            try:
                hue = camera.BslHueValue.value
            except AttributeError:
                hue = NotSupported()
            except AccessException:
                hue = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        saturation: Union[float, NotAvailable]
        try:
            saturation = camera.Saturation.value
        except (AttributeError, AccessException):
            try:
                saturation = camera.BslSaturationValue.value
            except AttributeError:
                saturation = NotSupported()
            except AccessException:
                saturation = TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        return ColorAdjustment(hue_rotation=hue, saturation_adjustment=saturation, color_adjustments=color_adjustments)

    def set_color_adjustment(
        self, sensor: Camera, color_adjustment: Union[ColorAdjustment, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the color adjustment.

        Arguments:
            sensor (Camera): The sensor to which apply the configuration.

            color_adjustment (ColorAdjustment): the new color adjustment.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        if not is_available(color_adjustment):
            return
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        camera = node_map

        for color, hs in color_adjustment.color_adjustments.items():
            try:
                if is_available(hs):
                    camera.ColorAdjustmentSelector.value = color
                    camera.ColorAdjustmentHue.value = hs.hue_rotation
                    camera.ColorAdjustmentSaturation.value = hs.saturation_adjustment
            except AttributeError as e:
                raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
            except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                raise SensorLost(sensor)

        hue = color_adjustment.hue_rotation if is_available(color_adjustment.hue_rotation) else 0
        try:
            camera.Hue.value = hue
        except (AttributeError, AccessException):
            try:
                camera.BslHueValue.value = hue
            except AttributeError as e:
                if is_available(color_adjustment.hue_rotation):
                    raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        saturation = (
            color_adjustment.saturation_adjustment if is_available(color_adjustment.saturation_adjustment) else 1
        )
        try:
            camera.Saturation.value = saturation
        except (AttributeError, AccessException):
            try:
                camera.BslSaturationValue.value = saturation
            except AttributeError as e:
                if is_available(color_adjustment.saturation_adjustment):
                    raise SensorFeaturePermanentlyNotAvailable(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        sensor.color_adjustment = color_adjustment

    #
    # Device temperature
    #

    def fetch_device_temperature(self, sensor: Camera) -> Union[float, NotAvailable]:
        """Return the current device temperature.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's temperature

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            DriverFeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        camera = node_map
        try:
            return camera.DeviceTemperature.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def fetch_device_temperature_state(self, sensor: Camera) -> Union[TemperatureState, NotAvailable]:
        """Return the current device temperature state.

        The Temperature State camera indicates whether the camera's internal temperature is normal or too high.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's temperature state.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            DriverFeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        camera = node_map
        try:
            value = camera.TemperatureState.value
            return {
                'Critical': TemperatureState.CRITICAL,
                'Ok': TemperatureState.OKAY,
                'Error': TemperatureState.ERROR,
            }[value]
        except AttributeError:
            return NotSupported()
        except (AccessException, KeyError):
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    #
    # Pixel format
    #

    def fetch_pixel_format(self, sensor: Camera) -> Union[PixelFormat, NotAvailable]:
        """Return the current device pixel format.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's pixel format.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            name = node_map.PixelFormat.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            return PixelFormat.from_name(name)
        except FeatureUnavailable as e:
            logger.warning(f'Pixel format not implemented {e}')
            return NotSupported()

    def set_pixel_format(
        self, sensor: Camera, pixel_format: Union[PixelFormat, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Apply the pixel format.

        Arguments:
            sensor (Camera): The sensor to which apply the configuration.

            pixel_format (PixelFormat): the new pixel format.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        if not is_available(pixel_format):
            return

        pixel_format_str: str = pixel_format.designation()

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            node_map.PixelFormat.value = pixel_format_str
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        sensor.pixel_format = pixel_format_str

    #
    # Crop
    #

    def fetch_crop(self, sensor: Camera) -> Union[Roi, NotAvailable]:
        """Return the current crop.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[Roi]: The sensor's crop.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            w = node_map.Width.value
            h = node_map.Height.value
            x = node_map.OffsetX.value
            y = node_map.OffsetY.value
            return Roi(x=x, y=y, width=w, height=h)
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_crop(
        self, sensor: Camera, crop: Optional[Union[Roi, NotAvailable]] = None, bounds_check: bool = False
    ) -> None:
        """Apply the crop.

        Arguments:
            sensor (Camera): The sensor to which apply the configuration.

            crop (Roi): the new crop.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        if not is_available(crop) and crop is not None:
            return

        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        x, y, w, h = 0, 0, None, None
        if crop is not None:
            x, y, w, h = crop.x, crop.y, crop.width, crop.height

        try:
            previous_x = node_map.OffsetX.value
            previous_y = node_map.OffsetY.value
            previous_w = node_map.Width.value
            previous_h = node_map.Height.value
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        if w is None or h is None:
            try:
                w = node_map.WidthMax.value
                h = node_map.HeightMax.value
            except AttributeError:
                try:
                    w = previous_x + node_map.Width.max
                    h = previous_y + node_map.Height.max
                except AttributeError:
                    raise SensorFeaturePermanentlyNotAvailable
                except AccessException as e:
                    raise SensorFeatureTemporarilyNotAvailable(e)
                except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                    raise SensorLost(sensor)

        assert w is not None and h is not None

        # @dev: If not resetting offsetX or Width to their minimal values, an exception can
        #       occur even if the pair (x, width) is valid.
        #       Example: cfg 1: x, width = (1, max_width -1)
        #                cfg 2: x, width = (0, max_width)
        #                cfg 1 -> cfg 2: raises if we set width first.
        #                cfg 2 -> cfg 1: raises if we set x first.
        #       Therefore, we set the value to the min value (assumed to be zero) and apply the
        #       pair afterwards.
        try:
            if node_map.OffsetX.value != 0:
                node_map.OffsetX.value = 0
            if node_map.Width.value != w:
                node_map.Width.value = w
            if node_map.OffsetX.value != x:
                node_map.OffsetX.value = x
        except OutOfRangeException as e:
            # Thrown if the width is not dividable by the increment, or w out of range.
            try:
                node_map.OffsetX.value = previous_x
                node_map.Width.value = previous_w
            except (AttributeError, OutOfRangeException):
                raise InvalidArgumentException(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        try:
            if node_map.OffsetY.value != 0:
                node_map.OffsetY.value = 0

            if node_map.Height.value != h:
                node_map.Height.value = h

            if node_map.OffsetY.value != y:
                node_map.OffsetY.value = y
        except OutOfRangeException as e:
            # Thrown if the width is not dividable by the increment, or h out of range.
            try:
                node_map.OffsetY.value = previous_y
                node_map.Height.value = previous_h
            except (AttributeError, OutOfRangeException):
                raise InvalidArgumentException(e)
            except AccessException as e:
                raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        sensor.crop = crop

    #
    # Camera configurations
    #

    def fetch_current_camera_configuration_identifier(self, sensor: Camera) -> Union[str, NotAvailable]:
        """Return the current camera configuration. This does not convey any information on
        whether the current settings are actually loaded from the camera configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[str]: The sensor's camera configuration name.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            return node_map.UserSetSelector.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_current_camera_configuration_identifier(
        self, sensor: Camera, value: Union[str, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Sent the current camera configuration to the camera configuration that is identified
        by the provided name. This does not automatically load the configuration, as we might
        want to set the current camera configuration just to apply our current configuration to it.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            value (str): the camera configuration to apply / load.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        if not is_available(value):
            return
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)

        try:
            node_map.UserSetSelector.value = value
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

        sensor.current_camera_configuration_identifier = value

    def load_current_camera_configuration(self, sensor: Camera) -> None:
        """Load the current configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        was_started = sensor.state == SensorState.STARTED
        if was_started:
            self.stop(sensor)
        try:
            node_map.UserSetLoad.execute()
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)
        finally:
            if was_started:
                self.start(sensor)

    def save_current_camera_configuration(self, sensor: Camera) -> None:
        """Save the current configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            node_map.UserSetSave.execute()
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def fetch_startup_camera_configuration_name(self, sensor: Camera) -> Union[str, NotAvailable]:
        """Return the startup camera configuration name.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (Camera): The sensor from which we gather information.

        Returns:
            Optional[str]: The sensor's camera configuration name.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            return node_map.UserSetDefault.value
        except AttributeError:
            return NotSupported()
        except AccessException:
            return TemporarilyNotAvailable()
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def set_startup_camera_configuration_name(
        self,
        sensor: Camera,
        startup_camera_configuration_name: Union[str, NotAvailable],
        bounds_check: bool = False,
    ) -> None:
        """Sent the startup camera configuration to the camera configuration that is identified
        by the provided name. This does not automatically load the configuration, as we might
        want to set the current camera configuration just to apply our current configuration to it.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            startup_camera_configuration_name (str): the camera configuration to apply / load.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        if not is_available(startup_camera_configuration_name):
            return
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        try:
            node_map.UserSetDefault.value = startup_camera_configuration_name
        except AttributeError as e:
            raise SensorFeaturePermanentlyNotAvailable(e)
        except AccessException as e:
            raise SensorFeatureTemporarilyNotAvailable(e)
        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def export_configuration(self, sensor: Camera) -> bytes:
        """Dump the configuration to a bytes array.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

        Return:
            bytes: Configuration dump.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        sensor = self._sensor(sensor)
        node_map = self._node_map(sensor)
        started: bool = sensor.state == SensorState.STARTED
        if started:
            self.stop(sensor)

        genicam_settings: bytes = dump_genicam_settings(node_map.nodes)

        if started:
            self.start(sensor)

        return genicam_settings

    def import_configuration(self, sensor: Camera, configuration: bytes):
        """Read the configuration from a previous configuration dump.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

            configuration (bytes) : Configuration dump.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorInvalidState if the sensor's state is not CONNECTED or STARTED.

            DriverStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        sensor = self._sensor(sensor)
        image_acquirer = self._image_acquirer(sensor)
        node_map = self._node_map(sensor)
        started: bool = sensor.state == SensorState.STARTED

        if started:
            self.stop(sensor)

        load_genicam_settings(image_acquirer, node_map.nodes, configuration)

        if started:
            self.start(sensor)

    @staticmethod
    def camel_to_snake(name) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()

    # @abstractmethod
    def fetch_sensor_data(self, sensor: Camera, timeout: float = 0.0) -> Union[CameraData, NotAvailable]:
        sensor = self._sensor(sensor)
        image_acquirer = self._image_acquirer(sensor, {SensorState.STARTED})
        camera = self._node_map(sensor, {SensorState.STARTED})

        try:
            with image_acquirer.fetch(timeout=timeout) as buffer:
                if any(True for c in buffer.payload.components if c.delivered_image_height != c.height):
                    raise IncompleteBufferError('Partial buffer arrived!')

                values = {}
                try:
                    for key in camera.ChunkSelector.symbolics:
                        try:
                            values[self.camel_to_snake(key)] = camera.__getattr__(f'Chunk{key}').value
                        except (AttributeError, AccessException):
                            pass
                except (AttributeError, AccessException):
                    pass

                components = []
                for component in buffer.payload.components:
                    data_dict = {'data': component.data.copy()}
                    for field in fields(Component):
                        try:
                            data_dict[field.name] = getattr(component, field.name, None)
                        except (AttributeError, genicam.gentl.NotImplementedException):
                            data_dict[field.name] = None

                    components.append(Component(**data_dict))

                data_dict = {'components': components}
                for field in fields(ImageContent):
                    if field.name != 'components':
                        try:
                            data_dict[field.name] = getattr(buffer, field.name, None)
                        except (AttributeError, genicam.gentl.NotImplementedException):
                            data_dict[field.name] = None

                content = ImageContent(**data_dict)

                image_data = []
                for component in content.components:
                    component.data.shape = (component.height, component.width, -1)
                    image_data.append(component.data)

                cd = CameraData(timestamp_inference=content.timestamp_ms, content=image_data)

                cd.meta_information = copy.copy(values)
                for k, v in values.items():
                    if hasattr(cd, k):
                        if not getattr(cd, k):
                            setattr(cd, k, v)
                            del cd.meta_information[k]

                sensor.acquisition_timestamp = int(time.time() * 1000)
                return cd
        except TimeoutException:
            return TemporarilyNotAvailable()

        except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
            raise SensorLost(sensor)

    def ping(self):
        for s in self.sensors.values():
            if s.state in {SensorState.STARTED, SensorState.CONNECTED}:
                node_map = self._node_map(s)
                try:
                    node_map.PixelFormat.value
                except AttributeError:
                    return NotSupported()
                except AccessException:
                    return TemporarilyNotAvailable()
                except (NotAvailableException, ErrorException, ResourceInUseException, ClosedException):
                    raise SensorLost(s)
